# Copyright 2022 MosaicML Composer authors
# SPDX-License-Identifier: Apache-2.0

"""Abstract class for utilities that upload to and download from object stores."""

import abc
import pathlib
from types import TracebackType
from typing import Callable, Optional, Type, Union

__all__ = ['ObjectStore', 'ObjectStoreTransientError']


class ObjectStoreTransientError(RuntimeError):
    """Custom exception class to signify transient errors.

    Implementations of the :class:`.ObjectStore` should re-raise any transient exceptions
    (e.g. too many requests, temporarily unavailable) with this class, so callers can easily
    detect whether they should attempt to retry any operation.

    For example, the :class:`.S3ObjectStore` does the following:

    .. testcode::

        from composer.utils import ObjectStore, ObjectStoreTransientError
        import botocore.exceptions

        class S3ObjectStore(ObjectStore):

            def upload_object(self, file_path: str, object_name: str):
                try:
                    ...
                except botocore.exceptions.ClientError as e:
                    if e.response['Error']['Code'] == 'LimitExceededException':
                        raise ObjectStoreTransientError(e.response['Error']['Code']) from e
                    raise e

    Then, callers can automatically handle exceptions:

    .. testcode::

        import time
        from composer.utils import ObjectStore, ObjectStoreTransientError

        def upload_file(object_store: ObjectStore, max_num_attempts: int = 3):
            for i in range(max_num_attempts):
                try:
                    object_store.upload_object(...)
                except ObjectStoreTransientError:
                    if i + 1 == max_num_attempts:
                        raise
                    else:
                        # Try again after exponential back-off
                        time.sleep(2**i)
                else:
                    # upload successful
                    return
    """
    pass


class ObjectStore(abc.ABC):
    """Abstract class for implementing object stores, such as LibcloudObjectStore and S3ObjectStore."""

    def get_uri(self, object_name: str) -> str:
        """Returns the URI for ``object_name``.

        .. note::

            This function does not check that ``object_name`` is in the object store.
            It computes the URI statically.

        Args:
            object_name (str): The object name.

        Returns:
            str: The URI for ``object_name`` in the object store.
        """
        raise NotImplementedError(f'{type(self).__name__}.get_uri is not implemented')

    def upload_object(
        self,
        object_name: str,
        filename: Union[str, pathlib.Path],
        callback: Optional[Callable[[int, int], None]] = None,
        **kwargs,
    ) -> None:
        """Upload an object currently located on a disk.

        Args:
            object_name (str): Object name (where object will be stored in the container)
            filename (str | pathlib.Path): Path to the object on disk
            callback ((int, int) -> None, optional): If specified, the callback is periodically called with the number of bytes
                uploaded and the total size of the object being uploaded.
            **kwargs: other arguments to the upload object function are supported
                and will be passed in to the underlying object store upload call.
                Currently only used for S3ObjectStore.

        Raises:
            ObjectStoreTransientError: If there was a transient connection issue with uploading the object.
        """
        del object_name, filename, callback, kwargs  # unused
        raise NotImplementedError(f'{type(self).__name__}.upload_object is not implemented')

    def get_object_size(self, object_name: str) -> int:
        """Get the size of an object, in bytes.

        Args:
            object_name (str): The name of the object.

        Returns:
            int: The object size, in bytes.

        Raises:
            FileNotFoundError: If the file was not found in the object store.
            ObjectStoreTransientError: If there was a transient connection issue with getting the object size.
        """
        raise NotImplementedError(f'{type(self).__name__}.get_object_size is not implemented')

    def download_object(
        self,
        object_name: str,
        filename: Union[str, pathlib.Path],
        overwrite: bool = False,
        callback: Optional[Callable[[int, int], None]] = None,
    ) -> None:
        """Download an object to the specified destination path.

        Args:
            object_name (str): The name of the object to download.
            filename (str | pathlib.Path): Full path to a file or a directory where the incoming file will be saved.
            overwrite (bool, optional): Whether to overwrite an existing file at ``filename``, if it exists.
                (default: ``False``)
            callback ((int) -> None, optional): If specified, the callback is periodically called with the number of bytes already
                downloaded and the total size of the object.

        Raises:
            FileExistsError: If ``filename`` already exists and ``overwrite`` is ``False``.
            FileNotFoundError: If the file was not found in the object store.
            ObjectStoreTransientError: If there was a transient connection issue with downloading the object.
        """
        del object_name, filename, overwrite, callback  # unused
        raise NotImplementedError(f'{type(self).__name__}.download_object is not implemented')

    def list_objects(self, prefix: Optional[str]) -> list[str]:
        """List all objects in the object store with the given prefix.

        Args:
            prefix (str): The prefix to search for.

        Returns:
            list[str]: A list of object names that match the prefix.
        """
        del prefix  # unused
        raise NotImplementedError(f'{type(self).__name__}.list_objects is not implemented')

    def close(self):
        """Close the object store."""
        pass

    def __enter__(self):
        return self

    def __exit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc: Optional[BaseException],
        traceback: Optional[TracebackType],
    ):
        del exc_type, exc, traceback  # unused
        self.close()



# Copyright 2022 MosaicML Composer authors
# SPDX-License-Identifier: Apache-2.0

"""Replaces all instances of :class:`torch.nn.LayerNorm` with a low precision :class:`torch.nn.LayerNorm` (either float16 or bfloat16).
By default, torch.autocast always runs torch.nn.LayerNorm in float32, so this surgery forces a lower precision.
"""

from composer.algorithms.low_precision_layernorm.low_precision_layernorm import (
    LowPrecisionLayerNorm,
    apply_low_precision_layernorm,
)

__all__ = ['LowPrecisionLayerNorm', 'apply_low_precision_layernorm']



# Copyright 2022 MosaicML Composer authors
# SPDX-License-Identifier: Apache-2.0

"""Eval Client base class and implementations."""

from composer.utils.eval_client.eval_client import EvalClient
from composer.utils.eval_client.lambda_eval_client import LambdaEvalClient
from composer.utils.eval_client.local_eval_client import LocalEvalClient
from composer.utils.eval_client.mosaicml_lambda_eval_client import MosaicMLLambdaEvalClient

__all__ = [
    'EvalClient',
    'LambdaEvalClient',
    'LocalEvalClient',
    'MosaicMLLambdaEvalClient',
]



# Copyright 2022 MosaicML Composer authors
# SPDX-License-Identifier: Apache-2.0

"""A collection of common torchmetrics."""
from __future__ import annotations

from typing import Callable

import torch
from torch import Tensor
from torchmetrics import Metric
from torchmetrics.utilities.data import to_categorical

from composer.loss import soft_cross_entropy

__all__ = ['MIoU', 'Dice', 'CrossEntropy', 'LossMetric']


class MIoU(Metric):
    """Torchmetrics mean Intersection-over-Union (mIoU) implementation.

    IoU calculates the intersection area between the predicted class mask and the label class mask.
    The intersection is then divided by the area of the union of the predicted and label masks.
    This measures the quality of predicted class mask with respect to the label. The IoU for each
    class is then averaged and the final result is the mIoU score. Implementation is primarily
    based on `mmsegmentation <https://github.com/open-mmlab/mmsegmentation/blob/aa50358c71fe9c4cccdd2abe42433bdf702e757b/mmseg/core/evaluation/metrics.py#L132>`_

    Args:
        num_classes (int): the number of classes in the segmentation task.
        ignore_index (int, optional): the index to ignore when computing mIoU. Default: ``-1``.
    """

    # Make torchmetrics call update only once
    full_state_update = False

    def __init__(self, num_classes: int, ignore_index: int = -1):
        super().__init__(dist_sync_on_step=True)
        self.num_classes = num_classes
        self.ignore_index = ignore_index
        self.add_state('total_intersect', default=torch.zeros(num_classes, dtype=torch.float64), dist_reduce_fx='sum')
        self.add_state('total_union', default=torch.zeros(num_classes, dtype=torch.float64), dist_reduce_fx='sum')

    def update(self, logits: Tensor, targets: Tensor):
        """Update the state with new predictions and targets."""
        preds = logits.argmax(dim=1)
        for pred, target in zip(preds, targets):
            mask = (target != self.ignore_index)
            pred = pred[mask]
            target = target[mask]

            intersect = pred[pred == target]
            area_intersect = torch.histc(intersect.float(), bins=self.num_classes, min=0, max=self.num_classes - 1)
            area_prediction = torch.histc(pred.float(), bins=self.num_classes, min=0, max=self.num_classes - 1)
            area_target = torch.histc(target.float(), bins=self.num_classes, min=0, max=self.num_classes - 1)

            self.total_intersect += area_intersect
            self.total_union += area_prediction + area_target - area_intersect

    def compute(self):
        """Aggregate state across all processes and compute final metric."""
        total_intersect = self.total_intersect[self.total_union != 0]  # type: ignore (third-party)
        total_union = self.total_union[self.total_union != 0]  # type: ignore (third-party)
        return 100 * (total_intersect / total_union).mean()


class Dice(Metric):
    """The Dice Coefficient for evaluating image segmentation.

    The Dice Coefficient measures how similar predictions and targets are.
    More concretely, it is computed as 2 * the Area of Overlap divided by
    the total number of pixels in both images.

    Args:
        num_classes (int): the number of classes in the segmentation task.
    """

    # Make torchmetrics call update only once
    full_state_update = False

    def __init__(self, num_classes: int):
        super().__init__(dist_sync_on_step=True)
        self.add_state('n_updates', default=torch.zeros(1), dist_reduce_fx='sum')
        self.add_state('dice', default=torch.zeros((num_classes,)), dist_reduce_fx='sum')

    def update(self, preds: Tensor, targets: Tensor):
        """Update the state based on new predictions and targets."""
        self.n_updates += 1  # type: ignore
        self.dice += self.compute_stats(preds, targets)

    def compute(self):
        """Aggregate the state over all processes to compute the metric."""
        dice = 100 * self.dice / self.n_updates  # type: ignore
        best_sum_dice = dice[:]
        top_dice = round(torch.mean(best_sum_dice).item(), 2)
        return top_dice

    @staticmethod
    def compute_stats(preds: Tensor, targets: Tensor):
        num_classes = preds.shape[1]
        scores = torch.zeros(num_classes - 1, device=preds.device, dtype=torch.float32)
        for i in range(1, num_classes):
            if (targets != i).all():
                # no foreground class
                _, _pred = torch.max(preds, 1)
                scores[i - 1] += 1 if (_pred != i).all() else 0
                continue
            _tp, _fp, _tn, _fn, _ = _stat_scores(preds, targets, class_index=i)  # type: ignore
            denom = (2 * _tp + _fp + _fn).to(torch.float)
            score_cls = (2 * _tp).to(torch.float) / denom if torch.is_nonzero(denom) else 0.0
            scores[i - 1] += score_cls
        return scores


def _stat_scores(
    preds: Tensor,
    targets: Tensor,
    class_index: int,
    argmax_dim: int = 1,
) -> tuple[Tensor, Tensor, Tensor, Tensor, Tensor]:

    if preds.ndim == targets.ndim + 1:
        preds = to_categorical(preds, argmax_dim=argmax_dim)

    tp = ((preds == class_index) * (targets == class_index)).to(torch.long).sum()
    fp = ((preds == class_index) * (targets != class_index)).to(torch.long).sum()
    tn = ((preds != class_index) * (targets != class_index)).to(torch.long).sum()
    fn = ((preds != class_index) * (targets == class_index)).to(torch.long).sum()
    sup = (targets == class_index).to(torch.long).sum()

    return tp, fp, tn, fn, sup


class CrossEntropy(Metric):
    """Torchmetrics cross entropy loss implementation.

    This class implements cross entropy loss as a :class:`torchmetrics.Metric` so that it can be returned by the
    :meth:`~.ComposerModel.metrics`.

    Args:
        ignore_index (int, optional): Specifies a target value that is ignored
            and does not contribute to the input gradient. ``ignore_index`` is only applicable when the target
            contains class indices. Default: ``-100``.

        dist_sync_on_step (bool, optional): sync distributed metrics every step. Default: ``False``.
    """

    # Make torchmetrics call update only once
    full_state_update = False

    def __init__(self, ignore_index: int = -100, dist_sync_on_step: bool = False):
        super().__init__(dist_sync_on_step=dist_sync_on_step)
        self.ignore_index = ignore_index
        self.add_state('sum_loss', default=torch.tensor(0.), dist_reduce_fx='sum')
        self.add_state('total_batches', default=torch.tensor(0), dist_reduce_fx='sum')

    def update(self, preds: Tensor, targets: Tensor) -> None:
        """Update the state with new predictions and targets."""
        # Loss calculated over samples/batch, accumulate loss over all batches
        self.sum_loss += soft_cross_entropy(preds, targets, ignore_index=self.ignore_index)
        assert isinstance(self.total_batches, Tensor)
        self.total_batches += 1

    def compute(self) -> Tensor:
        """Aggregate state over all processes and compute the metric."""
        # Return average loss over entire validation dataset
        assert isinstance(self.total_batches, Tensor)
        assert isinstance(self.sum_loss, Tensor)
        return self.sum_loss / self.total_batches


class LossMetric(Metric):
    """Turns a torch.nn Loss Module into distributed torchmetrics Metric.

    Args:
        loss_function (callable): loss function to compute and track.

        dist_sync_on_step (bool, optional): sync distributed metrics every step. Default: ``False``.
    """

    # Make torchmetrics call update only once
    full_state_update = False

    def __init__(self, loss_function: Callable, dist_sync_on_step: bool = False):
        super().__init__(dist_sync_on_step=dist_sync_on_step)
        self.loss_function = loss_function
        self.add_state('sum_loss', default=torch.tensor(0.), dist_reduce_fx='sum')
        self.add_state('total_batches', default=torch.tensor(0), dist_reduce_fx='sum')

    def update(self, preds: Tensor, targets: Tensor) -> None:
        """Update the state with new predictions and targets."""
        # Loss calculated over samples/batch, accumulate loss over all batches
        self.sum_loss += self.loss_function(preds, targets)
        self.total_batches += 1  # type: ignore

    def compute(self):
        """Aggregate state over all processes and compute the metric."""
        # Return average loss over entire validation dataset
        assert isinstance(self.total_batches, Tensor)
        assert isinstance(self.sum_loss, Tensor)
        return self.sum_loss / self.total_batches



# Copyright 2022 MosaicML Composer authors
# SPDX-License-Identifier: Apache-2.0

"""Helpers for running distributed data parallel training."""

import collections
import logging
import warnings
from contextlib import contextmanager, nullcontext
from typing import Any, Callable, ContextManager, Iterator, Optional, Sequence, Union, cast

import torch
from packaging import version
from torch.distributed.algorithms._checkpoint.checkpoint_wrapper import (
    CheckpointImpl,
    apply_activation_checkpointing,
    checkpoint_wrapper,
)
from torch.distributed.fsdp import FullyShardedDataParallel, ShardingStrategy
from torch.distributed.fsdp._common_utils import clean_tensor_name
from torch.nn.parallel import DistributedDataParallel
from torchmetrics import Metric, MetricCollection

from composer.core import Precision, State
from composer.devices import Device
from composer.distributed.meta_safe_apply import meta_safe_apply
from composer.distributed.mosaic_parallelism import (
    BACKWARD_PREFETCH_MAP,
    SHARDING_MAP,
    get_cpu_offload,
    get_mixed_precision,
    set_custom_fsdp_module_kwargs,
)
from composer.utils import FSDPConfig, StringEnum, TPConfig, dist, ensure_tuple

__all__ = ['DDPSyncStrategy', 'ddp_sync_context', 'prepare_ddp_module', 'prepare_fsdp_module', 'prepare_tp_module']

log = logging.getLogger(__name__)

process_group_cache = {}


class DDPSyncStrategy(StringEnum):
    """How and when gradient synchronization should happen.

    Attributes:
        SINGLE_AUTO_SYNC: The default behavior. Gradients are synchronized as they
            computed, for only the final microbatch of a batch. This is the most efficient
            strategy, but can lead to errors when ``find_unused_parameters`` is set, since
            it is possible different microbatches may use different sets of parameters,
            leading to an incomplete sync.
        MULTI_AUTO_SYNC: The default behavior when ``find_unused_parameters`` is set.
            Gradients are synchronized as they are computed for all microbatches. This ensures
            complete synchronization, but is less efficient than :attr:`SINGLE_AUTO_SYNC`. This
            efficiency gap is usually small, as long as either DDP syncs are a small portion
            of the trainer's overall runtime, or the number of microbatches per batch is
            relatively small.
        FORCED_SYNC: Gradients are manually synchronized only after all gradients have been
            computed for the final microbatch of a batch. Like :attr:`MULTI_AUTO_SYNC`, this
            strategy ensures complete gradient synchronization, but this tends to be slower than
            :attr:`MULTI_AUTO_SYNC`. This is because ordinarily syncs can happen in parallel
            with the ``loss.backward()`` computation, meaning syncs can be mostly complete by
            the time that function finishes. However, in certain circumstances, syncs may take
            a very long time to complete - if there are also a lot of microbatches per batch,
            this strategy may be optimal.
    """
    SINGLE_AUTO_SYNC = 'single_auto_sync'
    MULTI_AUTO_SYNC = 'multi_auto_sync'
    FORCED_SYNC = 'forced_sync'


@contextmanager
def ddp_sync_context(state: State, is_final_microbatch: bool, sync_strategy: Union[str, DDPSyncStrategy]):
    """A context manager for handling the :class:`DDPSyncStrategy`.

    Args:
        state (State): The state of the :class:`.Trainer`.
        is_final_microbatch (bool): Whether or not the context is being used during the final
            microbatch of the gradient accumulation steps.
        sync_strategy (str | DDPSyncStrategy): The ddp sync strategy to use. If a string
            is provided, the string must be one of the values in :class:`DDPSyncStrategy`.
    """
    if not isinstance(state.model, DistributedDataParallel):
        yield
        return

    assert state.optimizers is not None, 'optimizers have not been initialized'
    sync_strategy = DDPSyncStrategy(sync_strategy)

    no_sync_context = cast(Callable[[], ContextManager], state.model.no_sync)
    auto_sync_context = nullcontext

    if sync_strategy == DDPSyncStrategy.SINGLE_AUTO_SYNC:
        context = auto_sync_context if is_final_microbatch else no_sync_context
        with context():
            yield

    elif sync_strategy == DDPSyncStrategy.MULTI_AUTO_SYNC:
        with auto_sync_context():
            yield

    elif sync_strategy == DDPSyncStrategy.FORCED_SYNC:
        try:
            with no_sync_context():
                yield
        finally:
            if is_final_microbatch:
                for optimizer in state.optimizers:
                    for group in optimizer.param_groups:
                        for p in group['params']:
                            if p.grad is not None:
                                dist.all_reduce(p.grad)
                                p.grad = p.grad / dist.get_world_size()

    else:
        raise ValueError('Unknown sync strategy', sync_strategy)


def prepare_ddp_module(module: torch.nn.Module, find_unused_parameters: bool) -> torch.nn.Module:
    """Wraps the module in a :class:`torch.nn.parallel.DistributedDataParallel` object if running distributed training.

    Args:
        module (torch.nn.Module): The module to wrap.
        find_unused_parameters (bool): Whether or not to do a pass over the autograd graph
            to find parameters to not expect gradients for. This is useful if there are some
            parameters in the model that are not being trained.
    """
    if dist.is_available() and dist.is_initialized():
        if any((p.requires_grad for p in module.parameters())):
            log.debug('Wrapping model with DistributedDataParallel')
            ddp_model = DistributedDataParallel(module, find_unused_parameters=find_unused_parameters)
            return ddp_model
        return module
    if dist.is_available():
        raise RuntimeError('Please call dist.initialize_dist() before calling ddp.prepare_module()')

    raise RuntimeError(
        'When the world size is > 1, ``torch.distributed`` must be used. However, it is '
        'not available in your installation of PyTorch. Please install or build PyTorch '
        'with distributed support.',
    )


def _recreate_fsdp_param_groups_from_unwrapped_opt_info(
    fsdp_wrapped_named_params: Iterator[tuple[str, torch.nn.Parameter]],
    non_wrapped_param_names_to_group_num: dict[str, int],
    group_num_to_optimizer_info: dict[int, dict[str, Any]],
) -> list[dict[str, Any]]:
    """Helper function to recreate optimizer groups for FSDP wrapped modules.

    Optimizer param groups are formatted as:
    [
        {'params': [p1, p2], 'lr' : lr1}, # group 0
        {'params': [p3], 'lr' : lr2} # group 1
    ]
    ie. there are multiple parameters per group. Here, we track the group number in order to map
    multiple parameters to the same group

    Args:
        fsdp_wrapped_named_params: output of model.named_parameters() of an FSDP wrapped model
        non_wrapped_param_names_to_group_num: a dict mapping from the original model param names
                                            to the param group number
        group_num_to_optimizer_info: stores info like lr, eps for each group

    Returns a list of param groups, referencing the fsdp parameters
    """
    # Initialize an empty list of parameters for each optimizer group
    for group_num in group_num_to_optimizer_info.keys():
        group_num_to_optimizer_info[group_num]['params'] = []

    for fsdp_name, param in fsdp_wrapped_named_params:

        unwrapped_name = clean_tensor_name(fsdp_name)
        # need to have a 1:1 mapping between a fsdp param name and the non-wrapped vanilla param name
        retrieved_group_num = non_wrapped_param_names_to_group_num[unwrapped_name]
        group_num_to_optimizer_info[retrieved_group_num]['params'].append(param)

    # return sorted optimizer info groups
    return [group_num_to_optimizer_info[num] for num in sorted(group_num_to_optimizer_info.keys())]


def prepare_tp_module(
    model: torch.nn.Module,
    tp_config: TPConfig,
) -> None:
    """Prepare a module (assumed ComposerModel) for use with tensor parallel."""
    from torch.distributed.tensor.parallel import parallelize_module

    device_mesh = tp_config.device_mesh
    assert device_mesh is not None  # For type checking, set in State.__init__
    parallelize_module(
        module=model,
        device_mesh=device_mesh,
        parallelize_plan=tp_config.layer_plan,
    )


def prepare_fsdp_module(
    model: torch.nn.Module,
    optimizers: Optional[Union[torch.optim.Optimizer, Sequence[torch.optim.Optimizer]]],
    fsdp_config: FSDPConfig,
    precision: Precision,
    device: Device,
    auto_microbatching: bool,
    te_rng_seed: int = 1234,
) -> None:
    """Prepare a module (assumed ComposerModel) and optimizer for use with :class:`torch.distributed.fsdp.FullyShardedDataParallel`.

    Args:
        model (torch.nn.Module): The model to wrap.
        optimizers (torch.optim.Optimizer | Sequence[torch.optim.Optimizer], optional): The optimizer for `model`, assumed to have a single param group := model.parameters().
        fsdp_config (FSDPConfig): The FSDP config.
        precision: (Precision): The precision being used by the Trainer, used to fill in defaults for FSDP `mixed_precision` settings.
        device (Device): The device being used by the Trainer.
        auto_microbatching (bool, optional): Whether or not auto microbatching is enabled.
        te_rng_seed(int): The seed to use for the Transformer Engine activation checkpointing RNG. Defaults to 1234.
    """
    # Check sync_module_states is True for mixed initialization or HSDP
    if fsdp_config.sync_module_states == False:
        rank_on_meta = 1 if next(model.parameters()).device.type == 'meta' else 0
        all_ranks_meta = device.tensor_to_device(torch.tensor([rank_on_meta], dtype=torch.uint8))
        dist.all_reduce(all_ranks_meta, reduce_operation='MIN')
        any_ranks_meta = device.tensor_to_device(torch.tensor([rank_on_meta], dtype=torch.uint8))
        dist.all_reduce(any_ranks_meta, reduce_operation='MAX')
        if all_ranks_meta.item() == 0 and any_ranks_meta.item() == 1:
            raise ValueError(
                'Detected mixed initialization where some ranks have model on cpu or '
                'gpu and some ranks are on meta. Either keep all ranks on the same '
                "device or set parallelism_config['fsdp']['sync_module_states'] = True. Otherwise, "
                'some weights may be randomly initialized when loading a checkpoint.',
            )

    # Check if other ranks OOMed after forward/backward pass when using auto microbatching. This
    # may happen when close to memory limit or with uneven memory usage across ranks. Since we
    # need to do this before the model weights are gathered for the next FSDP block, we wrap every
    # FSPD block with a hook that checks if any other rank OOMed.
    def sync_hook(*args):
        # Check if any other rank hit an OOM
        found_cuda_oom_tensor = device.tensor_to_device(torch.tensor([0], dtype=torch.uint8))
        dist.all_reduce(found_cuda_oom_tensor, reduce_operation='MAX')
        found_cuda_oom = found_cuda_oom_tensor.item()
        # Signal current rank is still in batch
        all_ranks_finished_tensor = device.tensor_to_device(torch.tensor([0], dtype=torch.uint8))
        dist.all_reduce(all_ranks_finished_tensor, reduce_operation='MIN')

        if found_cuda_oom == 1:
            raise RuntimeError('CUDA out of memory encountered on a different rank')

    # Necessary variables for optimizers with multiple param groups in FSDP
    num_param_groups = None
    param_name_to_group_num = None
    group_num_to_param_group_info = None

    optimizer_specific_info = None
    if optimizers:
        optimizers_tuple = ensure_tuple(optimizers)
        if len(optimizers_tuple) != 1:
            raise NotImplementedError(f'Only one optimizer is supported; found {len(optimizers_tuple)} optimizers')

        # clearing optimizer param groups and state
        # that will be recreated at the end of prepare_fsdp_module
        optim = optimizers_tuple[0]

        num_param_groups = len(optim.param_groups)
        if num_param_groups > 1:
            if not fsdp_config.use_orig_params:
                raise RuntimeError(
                    'Multiple optimizer groups with FSDP are only supported with '
                    'use_orig_params=True.',
                )
            # optimizer.param_groups do not contain parameter names which are needed
            # to keep track of the different parameters in each group
            # so we use the pointers between model.parameters() and model.named_parameters()
            # to get the names of the parameters within optimizer.param_groups
            param_pointer_to_param_name = {id(p): n for n, p in model.named_parameters()}
            param_name_to_group_num = {}
            group_num_to_param_group_info = {}
            for group_num in range(len(optim.param_groups)):
                # Need to in-line to avoid a reference which causes FSDP to allocate extra GPU memory
                # group = optim.param_groups[group_num]
                for param_num in range(len(optim.param_groups[group_num]['params'])):
                    # Need to in-line to avoid a reference which causes FSDP to allocate extra GPU memory
                    # param = optim.param_groups[group_num]['params'][param_num]
                    param_name_to_group_num[param_pointer_to_param_name[id(
                        optim.param_groups[group_num]['params'][param_num],
                    )]] = group_num

                # this includes optimizer-specific values like lr, eps
                # this will be used as the kwargs for the optim param groups later
                optimizer_specific_group_info = {
                    k: v for k, v in optim.param_groups[group_num].items() if k != 'params'
                }
                group_num_to_param_group_info[group_num] = optimizer_specific_group_info
        else:
            optimizer_specific_info = {k: v for k, v in optim.param_groups[0].items() if k != 'params'}

        optim.param_groups.clear()
        optim.state.clear()

    sharding_map_key = fsdp_config.sharding_strategy.upper()
    sharding_strategy = SHARDING_MAP[sharding_map_key]

    kwargs = {}
    if version.parse(
        torch.__version__.split('.dev')[0],
    ) >= version.parse('2.2.0') and fsdp_config.device_mesh is not None:
        if fsdp_config.process_group is not None:
            warnings.warn(
                'process_group and device_mesh are set for FSDP, so ignoring device_mesh. Please set process_group to None.',
            )
        else:
            ndim = fsdp_config.device_mesh.ndim
            if ndim == 1 and sharding_strategy == ShardingStrategy.HYBRID_SHARD:
                sharding_strategy = ShardingStrategy.FULL_SHARD
                warnings.warn('HYBRID_SHARD is not supported with 1D device mesh. Using FULL_SHARD instead.')
            elif ndim == 1 and sharding_strategy == ShardingStrategy._HYBRID_SHARD_ZERO2:
                sharding_strategy = ShardingStrategy.SHARD_GRAD_OP
                warnings.warn('_HYBRID_SHARD_ZERO2 is not supported with 1D device mesh. Using SHARD_GRAD_OP instead.')
            elif ndim == 2 and sharding_strategy == ShardingStrategy.SHARD_GRAD_OP:
                sharding_strategy = ShardingStrategy._HYBRID_SHARD_ZERO2
                warnings.warn('SHARD_GRAD_OP is not supported with 2D device mesh. Using _HYBRID_SHARD_ZERO2 instead.')
            elif ndim == 2 and sharding_strategy == ShardingStrategy.FULL_SHARD:
                sharding_strategy = ShardingStrategy.HYBRID_SHARD
                warnings.warn('FULL_SHARD is not supported with 2D device mesh. Using HYBRID_SHARD instead.')
            kwargs['device_mesh'] = fsdp_config.device_mesh

    cpu_offload = get_cpu_offload(cpu_offload=fsdp_config.cpu_offload)

    mixed_precision = fsdp_config.mixed_precision
    keep_low_precision_grads = fsdp_config.keep_low_precision_grads
    mixed_precision, _, _, _ = get_mixed_precision(
        precision,
        mixed_precision=mixed_precision,
        keep_low_precision_grads=keep_low_precision_grads,
    )

    process_group = None
    if fsdp_config.process_group is not None:
        process_group_dict = {'process_group': fsdp_config.process_group}
        process_group = set_custom_fsdp_module_kwargs(process_group_dict, process_group_cache)['process_group']
    backward_prefetch = BACKWARD_PREFETCH_MAP[fsdp_config.backward_prefetch.upper()]
    activation_checkpointing = fsdp_config.activation_checkpointing
    activation_cpu_offload = fsdp_config.activation_cpu_offload
    sync_module_states = fsdp_config.sync_module_states
    forward_prefetch = fsdp_config.forward_prefetch
    limit_all_gathers = fsdp_config.limit_all_gathers
    ignored_modules = fsdp_config.ignored_modules
    state_dict_type = fsdp_config.state_dict_type
    activation_checkpointing_reentrant = fsdp_config.activation_checkpointing_reentrant
    te_checkpoint_wrapper = fsdp_config.te_checkpoint_wrapper if precision == Precision.AMP_FP8 else False
    te_shard_fp8_weight = fsdp_config.te_shard_fp8_weight if precision == Precision.AMP_FP8 else False
    sharded_ckpt_prefix_dir = fsdp_config.sharded_ckpt_prefix_dir
    use_orig_params = fsdp_config.use_orig_params

    # We choose to not wrap the ComposerModel directly, but instead wrap any submodules like `ComposerModel.model`
    # This makes it safer to call ComposerModel-specific functions like 'eval_forward' that
    # may make calls to sharded submodules. If we only wrap the submodules, then any call that ComposerModel makes
    # to a FSDP-wrapped submodule's `forward()` function will be safe and all-gather the necessary weights before `forward()`.
    for obj_name, obj in model.named_children():
        if not isinstance(obj, (Metric, MetricCollection)):

            # Skip wrapping submodules which are explicitly marked with no wrap
            if hasattr(obj, '_fsdp_wrap') and not bool(obj._fsdp_wrap):
                continue

            # Rather than verifying these changes with older PyTorch versions, we are fixing forward here
            if version.parse(torch.__version__) > version.parse('2.1.0'):
                # A dictionary of all tied parameter pointers to (module, attr) tuples
                tied_pointers = {}

                # Goes through all modules finding which weights have the same pointers
                for mod in obj.modules():
                    for attr_name, attr in mod.named_parameters(recurse=False):
                        ptr = id(attr)
                        mod_attr_list = tied_pointers.get(ptr, [])
                        mod_attr_list.append((mod, attr_name))
                        tied_pointers[ptr] = mod_attr_list

                # Dictionary mapping the source module to a list of (target module, source attr, target attr) tuples
                source_mod_to_mod_attr = {}
                for mod_attr_list in tied_pointers.values():
                    # If there is only one module for this pointer, then there is no weight tying
                    if len(mod_attr_list) == 1:
                        continue

                    # Arbitrarily choose the first module as the source module
                    first_mod, first_attr = mod_attr_list[0]
                    source_mod_to_mod_attr[first_mod] = [
                        (target_mod, first_attr, dest_attr) for target_mod, dest_attr in mod_attr_list[1:]
                    ]

                # Clean up no longer needed module references for memory safety
                del tied_pointers

                def _param_init_fn(module: torch.nn.Module) -> None:
                    # If we do not have any parameters or buffers on meta device managed by this module directly, we do not need to call the parameter init function.
                    # It is assumed that whatever process moved the parameters off of meta device initialized them.
                    # We expect this to occur if we have tied weights, as the second module will already have the weights initialized.
                    is_meta = any(param.is_meta for param in module.parameters(recurse=False)
                                 ) or any(buffer.is_meta for buffer in module.buffers(recurse=False))
                    if not is_meta:
                        return

                    # Move all parameters and buffers to the current device
                    module.to_empty(device=f'cuda:{torch.cuda.current_device()}', recurse=False)

                    # Redo weight tying, which will have been broken by the above line that moves parameters off of meta device
                    if module in source_mod_to_mod_attr:
                        for target_mod, first_attr, dest_attr in source_mod_to_mod_attr[module]:
                            setattr(target_mod, dest_attr, getattr(module, first_attr))

                    # Run the specified initialization
                    if hasattr(obj, 'param_init_fn') and isinstance(obj.param_init_fn, Callable):
                        obj.param_init_fn(module)
                    elif hasattr(module, 'reset_parameters') and isinstance(module.reset_parameters, Callable):
                        module.reset_parameters()
                    else:
                        raise ValueError(
                            f'Object `{obj_name}` does not have a ``param_init_fn`` or a ``reset_parameters`` function. '
                            'This leaves parameters without initialization. Please add a ``param_init_fn`` or ``reset_parameters`` '
                            f'to module `{obj_name}`.',
                        )
            else:

                def _param_init_fn(module: torch.nn.Module) -> None:
                    # A dictionary of all tied parameter pointers to module names
                    tied_pointers = {}

                    # Goes through all modules finding which weights have the same pointers
                    for name, mod in module.named_modules():
                        # Since FSDP recursively wraps, at parent modules we can encounter already
                        # wrapped weights, as a result we should skip any modules with `_fsdp_wrapped_module.`
                        if '_fsdp_wrapped_module' in name:
                            continue
                        for attr in ['weight', 'bias']:
                            if hasattr(mod, attr):
                                mod_attr = getattr(mod, attr)
                                if mod_attr is None:
                                    continue
                                ptr = id(mod_attr)
                                ptr_attr = (ptr, attr)
                                name_list = tied_pointers.get(ptr_attr, [])
                                name_list.append(name)
                                tied_pointers[ptr_attr] = name_list

                    # Creates a dictionary of module names that should be tied together
                    tied_mod_names = collections.defaultdict(list)
                    # Creates a set of modules we should not initialize
                    should_not_init_params = set()
                    for ptr_attr_type, mod_names in tied_pointers.items():
                        # No modules for this pointer are tied
                        if len(mod_names) == 1:
                            continue
                        _, attr_type = ptr_attr_type
                        first = next(mod_names.__iter__())
                        for elem in mod_names:
                            should_not_init_params.add('.'.join([elem, attr_type]))
                            tied_mod_names[(first, attr_type)].append(elem)
                        # Make sure at least one of the tied parameters is initialized
                        should_not_init_params.remove('.'.join([first, attr_type]))

                    meta_safe_apply(
                        module,
                        lambda t: torch.empty_like(t, device=f'cuda:{torch.cuda.current_device()}'),
                        should_not_init_params,
                        module_name='',
                    )

                    if len(tied_mod_names) > 0:
                        warnings.warn((
                            'The passed in model appears to have tied weights. In order to '
                            'support effective weight tying, the tied modules need to be '
                            'in the same FSDP module. If the weights are not properly tied '
                            'it can lead to loss spikes. We have tried our best to ensure '
                            'the tied weights are in the same FSDP module.'
                        ))

                    # Redoes weight tying
                    for name_attr, tied_names in tied_mod_names.items():
                        name, attr = name_attr
                        src_mod = module.get_submodule(name)
                        # We need to make sure the source and destination
                        # modules end up in the same FSDP module otherwise
                        # with sharding weight tying gets violated
                        src_mod._fsdp_wrap = False  # type: ignore
                        src_params = getattr(src_mod, attr)
                        for tied_name in tied_names:
                            dest_mod = module.get_submodule(tied_name)
                            dest_mod._fsdp_wrap = False  # type: ignore
                            setattr(dest_mod, attr, src_params)

                    if hasattr(obj, 'param_init_fn') and isinstance(obj.param_init_fn, Callable):
                        module.apply(obj.param_init_fn)
                    elif hasattr(module, 'reset_parameters') and isinstance(module.reset_parameters, Callable):
                        module.reset_parameters()
                    else:
                        raise ValueError(
                            f'Object `{obj_name}` does not have a ``param_init_fn`` or a ``reset_parameters`` function. '
                            'This leaves parameters without initialization. Please add a ``param_init_fn`` or ``reset_parameters`` '
                            f'to module `{obj_name}`.',
                        )

            if version.parse(torch.__version__) > version.parse('2.1.0.dev'):
                # CustomPolicy is only supported in torch v2.1.0-rc1 or higher
                from torch.distributed.fsdp.wrap import CustomPolicy  # type: ignore

                def lambda_fn(module: torch.nn.Module) -> Union[bool, dict]:
                    ret = False
                    if hasattr(module, '_fsdp_wrap'):
                        ret = bool(module._fsdp_wrap)
                    elif hasattr(obj, 'fsdp_wrap_fn') and isinstance(obj.fsdp_wrap_fn, Callable):
                        ret = obj.fsdp_wrap_fn(module)
                        if isinstance(ret, dict):
                            ret = set_custom_fsdp_module_kwargs(ret, process_group_cache)
                    if ret and auto_microbatching:
                        module.register_forward_hook(sync_hook)
                        module.register_full_backward_hook(sync_hook)
                    return ret

                _auto_wrap_policy = CustomPolicy(lambda_fn)
            else:
                # Choose which modules to FSDP wrap according to the following priority:
                # If module has attribute `module._fsdp_wrap = ...`, always respect it
                # Otherwise wrap if root object `obj.fsdp_wrap_fn(module)` is true.
                def __auto_wrap_policy(module: torch.nn.Module, recurse: bool, nonwrapped_numel: int) -> bool:
                    if recurse:
                        return True
                    should_be_wrapped = False
                    if hasattr(module, '_fsdp_wrap'):
                        should_be_wrapped = bool(module._fsdp_wrap)
                    elif hasattr(obj, 'fsdp_wrap_fn') and isinstance(obj.fsdp_wrap_fn, Callable):
                        should_be_wrapped = obj.fsdp_wrap_fn(module)

                    if should_be_wrapped and auto_microbatching:
                        module.register_forward_hook(sync_hook)
                        module.register_full_backward_hook(sync_hook)
                    return should_be_wrapped

                def _auto_wrap_policy_new(module: torch.nn.Module, recurse: bool, nonwrapped_numel: int) -> bool:
                    return __auto_wrap_policy(module, recurse, nonwrapped_numel)

                _auto_wrap_policy = _auto_wrap_policy_new

            fsdp_obj = FullyShardedDataParallel(
                obj,
                process_group=process_group,
                sharding_strategy=sharding_strategy,
                auto_wrap_policy=_auto_wrap_policy,  # type: ignore FSDP type bug
                cpu_offload=cpu_offload,
                mixed_precision=mixed_precision,
                backward_prefetch=backward_prefetch,
                ignored_modules=ignored_modules,
                param_init_fn=_param_init_fn,
                device_id=torch.cuda.current_device(),
                sync_module_states=sync_module_states,
                forward_prefetch=forward_prefetch,
                limit_all_gathers=limit_all_gathers,
                use_orig_params=use_orig_params,
                **kwargs,
            )

            if te_shard_fp8_weight:
                try:
                    from transformer_engine.pytorch.distributed import prepare_te_modules_for_fsdp
                except ModuleNotFoundError:
                    raise ModuleNotFoundError('Please install transformer-engine to use prepare_te_modules_for_fsdp')
                log.info(f'Calling prepare_te_modules_for_fsdp to enable TE weights sharding')
                prepare_te_modules_for_fsdp(fsdp_obj)

            if hasattr(fsdp_obj, '_exec_order_data'):
                if hasattr(fsdp_obj._exec_order_data, '_forward_prefetch_limit'):
                    fsdp_obj._exec_order_data._forward_prefetch_limit = fsdp_config.forward_prefetch_limit
                else:
                    warnings.warn(
                        'FSDP._exec_order_data does not have attribute _forward_prefetch_limit '
                        'which is unexpected and will result in `forward_prefetch_limit` from FSDP '
                        'config being ignored. Please open an issue to Composer to report this.',
                    )
                if hasattr(fsdp_obj._exec_order_data, '_backward_prefetch_limit'):
                    fsdp_obj._exec_order_data._backward_prefetch_limit = fsdp_config.backward_prefetch_limit
                else:
                    warnings.warn(
                        'FSDP._exec_order_data does not have attribute _backward_prefetch_limit '
                        'which is unexpected and will result in `backward_prefetch_limit` from FSDP '
                        'config being ignored. Please open an issue to Composer to report this.',
                    )
            else:
                warnings.warn(
                    'FSDP does not have attribute _exec_order_data which is unexpected and will '
                    'result in `forward_prefetch_limit` and `backward_prefetch_limit` from FSDP '
                    'config being ignored. Please open an issue to Composer to report this.',
                )

            # Activation Checkpointing
            if activation_checkpointing or activation_cpu_offload:
                # FP8 TE requires using the TE checkpoint function, FSDP activation checkpointing only works with TE non-reentrant checkpointing
                if te_checkpoint_wrapper:
                    assert not activation_checkpointing_reentrant, 'TE checkpoint only works with non-reentrant checkpointing'
                if version.parse(torch.__version__) > version.parse('2.1.0.dev'):
                    from torch.distributed.algorithms._checkpoint.checkpoint_wrapper import offload_wrapper
                    if not activation_checkpointing_reentrant:
                        if te_checkpoint_wrapper:
                            try:
                                import transformer_engine.pytorch as te
                            except ModuleNotFoundError:
                                raise ModuleNotFoundError(
                                    'Please install transformer-engine to use TE checkpoint wrapper',
                                )

                            # RNG state tracker for checkpointing
                            CUDA_RNG_STATES_TRACKER = te.distributed.CudaRNGStatesTracker()
                            CUDA_RNG_STATES_TRACKER.add('fsdp-rng', te_rng_seed)

                            def get_cuda_rng_tracker():
                                return CUDA_RNG_STATES_TRACKER

                            first_wrap_fn = lambda m: checkpoint_wrapper(
                                m,
                                context_fn=te.distributed.get_activation_recompute_contexts,
                                checkpoint_fn=te.distributed.checkpoint,
                                use_reentrant=False,
                                get_rng_state_tracker=get_cuda_rng_tracker,
                            )
                        else:
                            first_wrap_fn = lambda m: checkpoint_wrapper(
                                m,
                                checkpoint_impl=CheckpointImpl.NO_REENTRANT,
                            ) if activation_checkpointing else (lambda module: module)
                        second_wrap_fn = (
                            lambda module: offload_wrapper(
                                first_wrap_fn(module)
                                if activation_checkpointing else module,  # type: ignore reportGeneralTypeIssues
                            )
                        ) if activation_cpu_offload else first_wrap_fn
                    else:

                        first_wrap_fn = lambda m: checkpoint_wrapper(
                            m,
                            checkpoint_impl=CheckpointImpl.REENTRANT,
                        ) if activation_checkpointing else (lambda module: module)
                        second_wrap_fn = (
                            lambda module: offload_wrapper(
                                first_wrap_fn(module)
                                if activation_checkpointing else module,  # type: ignore reportGeneralTypeIssues
                            )
                        ) if activation_cpu_offload else first_wrap_fn
                else:
                    if not activation_checkpointing_reentrant:
                        first_wrap_fn = lambda m: checkpoint_wrapper(
                            m,
                            checkpoint_impl=CheckpointImpl.NO_REENTRANT,
                        ) if activation_checkpointing else (lambda module: module)
                        second_wrap_fn = (
                            lambda module: checkpoint_wrapper(
                                first_wrap_fn(module),  # type: ignore reportGeneralTypeIssues
                                checkpoint_impl=CheckpointImpl.NO_REENTRANT,
                                offload_to_cpu=True,
                            )
                        ) if activation_cpu_offload else first_wrap_fn
                    else:
                        first_wrap_fn = checkpoint_wrapper if activation_checkpointing else (lambda module: module)
                        second_wrap_fn = (
                            lambda module: checkpoint_wrapper(
                                first_wrap_fn(module),  # type: ignore reportGeneralTypeIssues
                                offload_to_cpu=True,
                            )
                        ) if activation_cpu_offload else first_wrap_fn

                # Choose which modules to activation checkpoint according to the following priority:
                # If module has attribute `module._activation_checkpointing = ...`, always respect it
                # Otherwise checkpoint if root object `obj.activation_checkpointing_fn(module)` is true
                def _check_fn(module: torch.nn.Module) -> bool:
                    if isinstance(module, FullyShardedDataParallel):
                        return False
                    if hasattr(module, '_activation_checkpointing'):
                        return bool(module._activation_checkpointing)
                    if hasattr(
                        obj,
                        'activation_checkpointing_fn',
                    ) and isinstance(obj.activation_checkpointing_fn, Callable):
                        return obj.activation_checkpointing_fn(module)
                    return False

                apply_activation_checkpointing(
                    fsdp_obj,
                    checkpoint_wrapper_fn=second_wrap_fn,  # type: ignore
                    check_fn=_check_fn,  # type: ignore
                )

            setattr(model, obj_name, fsdp_obj)

    # Print FSDP wrapped model and FSDP config if `verbose=True`
    if fsdp_config.verbose:
        log.info(f'FSDP: Wrapped model: {model}')
        log.info(f'FSDP: Using sharding_strategy={sharding_strategy}')
        log.info(f'FSDP: Using cpu_offload={cpu_offload}')
        log.info(f'FSDP: Using mixed_precision={mixed_precision}')
        log.info(f'FSDP: Using backward_prefetch={backward_prefetch}')
        log.info(f'FSDP: Using activation_checkpointing={activation_checkpointing}')
        log.info(f'FSDP: Using activation_cpu_offload={activation_cpu_offload}')
        log.info(f'FSDP: Using te_checkpoint_wrapper={te_checkpoint_wrapper}')
        log.info(f'FSDP: Using te_shard_fp8_weight={te_shard_fp8_weight}')
        log.info(f'FSDP: Using sync_module_states={sync_module_states}')
        log.info(f'FSDP: Using forward_prefetch={forward_prefetch}')
        log.info(f'FSDP: Using limit_all_gathers={limit_all_gathers}')
        log.info(f'FSDP: Using state_dict_type={state_dict_type}')
        log.info(f'FSDP: Using sharded_ckpt_prefix_dir={sharded_ckpt_prefix_dir}')

    # Rebuild optimizer now that parameters are sharded
    if optimizers:
        optim = ensure_tuple(optimizers)[0]
        optim.param_groups.clear()

        assert num_param_groups is not None
        if num_param_groups > 1:
            assert param_name_to_group_num is not None
            assert group_num_to_param_group_info is not None

            param_groups = _recreate_fsdp_param_groups_from_unwrapped_opt_info(
                model.named_parameters(),
                param_name_to_group_num,
                group_num_to_param_group_info,
            )
            for param_group in param_groups:
                optim.add_param_group(param_group)
        else:
            assert optimizer_specific_info is not None
            optimizer_specific_info.update({'params': list(model.parameters())})
            optim.add_param_group(optimizer_specific_info)



# Copyright 2022 MosaicML Composer authors
# SPDX-License-Identifier: Apache-2.0

import numpy as np
import pytest
import torch
from torch.nn import functional as F

from composer.loss import DiceLoss, soft_cross_entropy
from composer.loss.utils import _one_hot, ensure_targets_one_hot, infer_target_type


def generate_targets():
    return [
        # Binary classification
        torch.randint(low=0, high=2, size=(8,), dtype=torch.long),
        # Classification
        torch.randint(low=0, high=10, size=(8,), dtype=torch.long),
        # Segmentation
        torch.randint(low=0, high=2, size=(8, 5, 5), dtype=torch.long),
        torch.randint(low=0, high=10, size=(8, 5, 5), dtype=torch.long),
        # 3D inputs
        torch.randint(low=0, high=10, size=(8, 5, 7, 11), dtype=torch.long),
        torch.randint(low=0, high=10, size=(8, 5, 8, 11), dtype=torch.long),
    ]


@pytest.mark.parametrize('targets', generate_targets())
class TestOneHot():

    def test_one_hot(self, targets):
        composer_one_hot = _one_hot(targets)
        pytorch_one_hot = F.one_hot(targets)
        torch.testing.assert_close(composer_one_hot, pytorch_one_hot)

    def test_one_hot_num_classes(self, targets):
        num_classes = targets.max() + 1
        composer_one_hot = _one_hot(targets, num_classes=num_classes)
        pytorch_one_hot = F.one_hot(targets, num_classes=num_classes)
        torch.testing.assert_close(composer_one_hot, pytorch_one_hot)

    @pytest.mark.parametrize('dim', [1])
    def test_one_hot_dim(self, targets, dim):
        composer_one_hot = _one_hot(targets, dim=dim)
        pytorch_one_hot = F.one_hot(targets)
        # Move class dim to specified dim
        pytorch_one_hot = torch.movedim(pytorch_one_hot, source=-1, destination=dim).contiguous()
        torch.testing.assert_close(composer_one_hot, pytorch_one_hot)

    @pytest.mark.xfail(raises=ValueError)
    def test_one_hot_wrong_type(self, targets):
        targets = _one_hot(targets.float())

    @pytest.mark.xfail(raises=ValueError)
    def test_one_hot_wrong_classes(self, targets):
        targets = _one_hot(targets, num_classes=1)


def fake_input_target_pairs(input_shape):
    num_classes = input_shape[1]
    reduced_input_shape = list(input_shape)
    reduced_input_shape.pop(1)

    input = torch.randn(input_shape)
    targets_idx = torch.randint(low=-1, high=num_classes, size=reduced_input_shape)
    targets_one_hot = torch.zeros_like(input)
    for i, value in np.ndenumerate(targets_idx):
        i_expanded = list(i)
        if value >= 0:
            i_expanded.insert(1, value)
            targets_one_hot[tuple(i_expanded)] = 1.0
    return input, targets_idx, targets_one_hot


def xfail(val):
    """shorthand to mark xfail parameters."""
    return pytest.param(val, marks=pytest.mark.xfail)


def generate_tensor_pairs():
    return [
        # Binary classification
        fake_input_target_pairs((64, 2)),
        # Classification
        fake_input_target_pairs((64, 10)),
        # Segmentation
        fake_input_target_pairs((64, 2, 5, 5)),
        fake_input_target_pairs((64, 10, 5, 5)),
        # 3D inputs
        fake_input_target_pairs((64, 2, 5, 7, 11)),
        fake_input_target_pairs((64, 10, 5, 7, 11)),
    ]


@pytest.mark.filterwarnings(
    r'ignore:Negative label indices are being ignored in conversion to one-hot labels:UserWarning',
)
@pytest.mark.parametrize('tensors', generate_tensor_pairs())
def test_ensure_targets_one_hot(tensors):
    input, targets_idx, targets_one_hot = tensors
    targets_one_hot_test = ensure_targets_one_hot(input, targets_idx)
    torch.testing.assert_close(targets_one_hot, targets_one_hot_test, check_stride=False)


@pytest.mark.parametrize('tensors', generate_tensor_pairs())
class TestSoftCrossEntropy:

    def test_infer_target_type(self, tensors):
        input, target_indices, target_onehot = tensors
        assert infer_target_type(input, target_indices) == 'indices'
        assert infer_target_type(input, target_onehot) == 'one_hot'

    # Note: Weights xfail as our implementation differs from PyTorch
    @pytest.mark.parametrize('reduction', ['mean', 'sum'])
    @pytest.mark.parametrize('use_weights', [xfail(True), False])
    @pytest.mark.filterwarnings(r'ignore:Some targets have less than 1 total probability:UserWarning')
    def test_soft_cross_entropy(self, tensors, use_weights, reduction):
        input, target_indices, target_onehot = tensors
        if use_weights:
            num_classes = target_onehot.shape[1]
            weights = torch.rand(size=[num_classes])
        else:
            weights = None

        loss_indices = soft_cross_entropy(input, target_indices, weight=weights, reduction=reduction, ignore_index=-1)
        loss_onehot = soft_cross_entropy(input, target_onehot, weight=weights, reduction=reduction)
        loss_reference = F.cross_entropy(input, target_indices, weight=weights, reduction=reduction, ignore_index=-1)

        torch.testing.assert_close(loss_indices, loss_onehot)
        torch.testing.assert_close(loss_indices, loss_reference)


@pytest.mark.filterwarnings('ignore::UserWarning')
@pytest.mark.parametrize('squared_pred', [True, False])
@pytest.mark.parametrize('jaccard', [True, False])
@pytest.mark.parametrize('batch', [True, False])
@pytest.mark.parametrize('ignore_absent_classes', [True, False])
class TestDiceLoss:

    @pytest.fixture()
    def target(self):
        target = torch.tensor([[[-1], [0], [1], [2]]]).repeat(1, 1, 4)
        target = torch.cat([target, target[:, [1, 2, 3, 0]], target[:, [2, 3, 0, 1]], target[:, [3, 0, 1, 2]]], dim=0)
        return target

    @pytest.fixture()
    def correct_input(self, target):
        input = target.clone()
        input[input == -1] = 1  # replace negative label with class prediction
        input = F.one_hot(input)
        input = torch.movedim(input, 3, 1)
        return input

    @pytest.fixture()
    def incorrect_input(self, correct_input):
        return correct_input[[3, 2, 1, 0]]

    @pytest.mark.parametrize('reduction', ['mean', 'sum'])
    def test_correct_prediction(
        self,
        correct_input: torch.Tensor,
        target: torch.Tensor,
        squared_pred: bool,
        jaccard: bool,
        batch: bool,
        ignore_absent_classes: bool,
        reduction: str,
    ):
        dice_loss = DiceLoss(
            squared_pred=squared_pred,
            jaccard=jaccard,
            batch=batch,
            ignore_absent_classes=ignore_absent_classes,
            reduction=reduction,
        )

        assert dice_loss(correct_input, target) == 0.0

    def test_incorrect_prediction(
        self,
        incorrect_input: torch.Tensor,
        target: torch.Tensor,
        squared_pred: bool,
        jaccard: bool,
        batch: bool,
        ignore_absent_classes: bool,
    ):
        dice_loss = DiceLoss(
            squared_pred=squared_pred,
            jaccard=jaccard,
            batch=batch,
            ignore_absent_classes=ignore_absent_classes,
        )
        loss = dice_loss(incorrect_input, target)
        torch.testing.assert_close(loss, torch.tensor(1.0))



# Copyright 2022 MosaicML Composer authors
# SPDX-License-Identifier: Apache-2.0

"""Util class to upload file."""

import logging
import multiprocessing
import os
import pathlib
import shutil
import tempfile
import time
import uuid
from concurrent.futures import Future, ProcessPoolExecutor
from typing import List

from composer.utils.dist import get_local_rank
from composer.utils.file_helpers import (
    maybe_create_object_store_from_uri,
)
from composer.utils.object_store.object_store import ObjectStore, ObjectStoreTransientError
from composer.utils.retrying import retry

log = logging.getLogger(__name__)

__all__ = ['RemoteUploader']


def _upload_file_to_object_store(
    remote_folder: str,
    remote_file_name: str,
    local_file_path: str,
    overwrite: bool,
    num_attempts: int,
) -> int:
    object_store: ObjectStore = maybe_create_object_store_from_uri(
        remote_folder,
    )  # pyright: ignore[reportGeneralTypeIssues]

    @retry(ObjectStoreTransientError, num_attempts=num_attempts)
    def upload_file(retry_index: int = 0):
        if retry_index == 0 and not overwrite:
            try:
                object_store.get_object_size(remote_file_name)
            except FileNotFoundError:
                # Good! It shouldn't exist.
                pass
            else:
                raise FileExistsError(f'Object {remote_file_name} already exists, but overwrite was set to False.')
        log.info(f'Uploading file {local_file_path} to {remote_file_name}')
        object_store.upload_object(
            object_name=remote_file_name,
            filename=local_file_path,
        )
        os.remove(local_file_path)

    log.info(f'Finished uploading file {local_file_path} to {remote_file_name}')
    # When encountering issues with too much concurrency in uploads, staggering the uploads can help.
    # This stagger is intended for use when uploading model shards from every rank, and will effectively reduce
    # the concurrency by a factor of num GPUs per node.
    local_rank = get_local_rank()
    local_rank_stagger = int(os.environ.get('COMPOSER_LOCAL_RANK_STAGGER_SECONDS', 0))
    log.debug(f'Staggering uploads by {local_rank * local_rank_stagger} seconds on {local_rank} local rank.')
    time.sleep(local_rank * local_rank_stagger)
    upload_file()
    return 0


class RemoteUploader:
    """Class for uploading a file to object store asynchronously."""

    def __init__(
        self,
        remote_folder: str,
        num_concurrent_uploads: int = 2,
        num_attempts: int = 3,
    ):
        if num_concurrent_uploads < 1 or num_attempts < 1:
            raise ValueError(
                f'num_concurrent_uploads and num_attempts must be >= 1, but got {num_concurrent_uploads} and {num_attempts}',
            )

        self.remote_folder = remote_folder
        # A folder to use for staging uploads
        self._tempdir = tempfile.TemporaryDirectory()
        self._upload_staging_folder = self._tempdir.name

        self.num_attempts = num_attempts

        self.executor = ProcessPoolExecutor(
            max_workers=num_concurrent_uploads,
            mp_context=multiprocessing.get_context('spawn'),
        )

        # Used internally to track the future status.
        # If a future completed successfully, we'll remove it from this list
        # when check_workers() or wait() is called
        self.futures: List[Future] = []

    def upload_file_async(
        self,
        remote_file_name: str,
        file_path: pathlib.Path,
        overwrite: bool,
    ):
        """Async call to submit a job for uploading.

        It returns a future, so users can track the status of the individual future.
        User can also call wait() to wait for all the futures.
        """
        # Copy file to staging folder
        copied_path = os.path.join(self._upload_staging_folder, str(uuid.uuid4()))
        os.makedirs(self._upload_staging_folder, exist_ok=True)
        shutil.copy2(file_path, copied_path)

        # Async upload file
        future = self.executor.submit(
            _upload_file_to_object_store,
            remote_folder=self.remote_folder,
            remote_file_name=remote_file_name,
            local_file_path=copied_path,
            overwrite=overwrite,
            num_attempts=self.num_attempts,
        )
        self.futures.append(future)
        return future

    def check_workers(self):
        """Non-blocking call to check workers are either running or done.

        Traverse self.futures, and check if it's completed
        1. if it completed with exception, raise that exception
        2. if it completed without exception, remove it from self.futures
        """
        done_futures: List[Future] = []
        for future in self.futures:
            if future.done():
                # future.exception is a blocking call
                exception_or_none = future.exception()
                if exception_or_none is not None:
                    raise exception_or_none
                else:
                    done_futures.append(future)
        for future in done_futures:
            self.futures.remove(future)

    def wait(self):
        """Blocking call to wait all the futures to complete.

        If a future is done successfully, remove it from self.futures(),
        otherwise, raise the exception
        """
        for future in self.futures:
            exception_or_none = future.exception()
            if exception_or_none is not None:
                raise exception_or_none
        self.futures = []

    def wait_and_close(self):
        """Blocking call to wait all uploading to finish and close this uploader.

        After this function is called, users can not use this uploader
        to uploading anymore. So please only call wait_and_close() after submitting
        all uploading requests.
        """
        # make sure all workers are either running, or completed successfully
        self.wait()
        self.executor.shutdown(wait=True)



# Copyright 2022 MosaicML Composer authors
# SPDX-License-Identifier: Apache-2.0

import glob
import inspect
import os
from urllib.parse import urlparse

import importlib_metadata
import pytest
import testbook
from testbook.client import TestbookNotebookClient

import composer
from composer.utils.import_helpers import MissingConditionalImportError
from tests.common import device

nb_root = os.path.join(os.path.dirname(composer.__file__), '..', 'examples')

NOTEBOOKS = [
    os.path.join(nb_root, nb) \
    for nb in glob.glob(os.path.join(nb_root, '*.ipynb')) \
]

try:
    importlib_metadata.files('mosaicml')
    package_name = 'mosaicml'
except importlib_metadata.PackageNotFoundError:
    try:
        importlib_metadata.files('composer')
        package_name = 'composer'
    except importlib_metadata.PackageNotFoundError:
        raise RuntimeError('Could not find the package under mosaicml or composer.')


def patch_notebooks():
    import itertools
    import multiprocessing

    from torch.utils.data import DataLoader

    from composer import Trainer

    multiprocessing.cpu_count = lambda: 2

    original_fit = Trainer.fit

    def new_fit(self: Trainer, *args, **kwargs):
        if 'duration' not in kwargs:
            kwargs['duration'] = '2ep'
        if 'train_subset_num_batches' not in kwargs:
            kwargs['train_subset_num_batches'] = 2
        if 'eval_dataloader' in kwargs and 'eval_subset_num_batches' not in kwargs:
            kwargs['eval_subset_num_batches'] = 1
        original_fit(self, *args, **kwargs)

    Trainer.fit = new_fit

    original_iter = DataLoader.__iter__

    def new_iter(self: DataLoader):
        return itertools.islice(original_iter(self), 2)

    DataLoader.__iter__ = new_iter  # type: ignore  # error: DataLoader has a stricter return type than islice


def modify_cell_source(tb: TestbookNotebookClient, notebook_name: str, cell_source: str, s3_bucket: str) -> str:
    # This function is called before each cell is executed
    if notebook_name == 'functional_api':
        # avoid div by 0 errors with batch size of 1
        cell_source = cell_source.replace('num_epochs = 5', 'num_epochs = 1')
        cell_source = cell_source.replace('acc_percent = 100 * num_right / eval_size', 'acc_percent = 1')
        cell_source = cell_source.replace('batch_size = 1024', 'batch_size = 64')
        cell_source = cell_source.replace('download=True', 'download=False')
    if notebook_name == 'custom_speedup_methods':
        cell_source = cell_source.replace('resnet_56', 'resnet_9')
        cell_source = cell_source.replace('batch_size=1024', 'batch_size=64')
        cell_source = cell_source.replace('download=True', 'download=False')
    if notebook_name == 'finetune_huggingface':
        cell_source = cell_source.replace(
            'sst2_dataset = datasets.load_dataset("glue", "sst2")',
            'sst2_dataset = datasets.load_dataset("glue", "sst2", download_mode="force_redownload")',
        )
        cell_source = cell_source.replace('batch_size=16', 'batch_size=2')
    if notebook_name == 'pretrain_finetune_huggingface':
        cell_source = cell_source.replace('batch_size=64', 'batch_size=1')
        cell_source = cell_source.replace('batch_size=32', 'batch_size=1')
    if notebook_name == 'early_stopping':
        cell_source = cell_source.replace('batch_size = 1024', 'batch_size = 64')
        cell_source = cell_source.replace('download=True', 'download=False')
    if notebook_name == 'getting_started':
        cell_source = cell_source.replace('batch_size = 1024', 'batch_size = 64')
        cell_source = cell_source.replace('download=True', 'download=False')
    if notebook_name == 'auto_microbatching':
        cell_source = cell_source.replace('batch_size = 2048', 'batch_size = 1024')
        cell_source = cell_source.replace('download=True', 'download=False')
    if notebook_name == 'migrate_from_ptl':
        cell_source = cell_source.replace('batch_size=256', 'batch_size=64')
        cell_source = cell_source.replace('download=True', 'download=False')

    cell_source = cell_source.replace("pip install 'mosaicml", f"pip install '{package_name}")
    cell_source = cell_source.replace('pip install mosaicml', f'pip install {package_name}')

    return cell_source


@pytest.mark.parametrize('notebook', NOTEBOOKS)
@device('cpu', 'gpu')
@pytest.mark.daily
def test_notebook(notebook: str, device: str, s3_bucket: str):
    trainer_monkeypatch_code = inspect.getsource(patch_notebooks)
    notebook_name = os.path.split(notebook)[-1][:-len('.ipynb')]

    if notebook_name == 'medical_image_segmentation':
        pytest.skip('Dataset is only available via kaggle; need to authenticate on ci/cd')
    if notebook_name == 'training_with_submitit':
        pytest.skip('The CI does not support SLURM and submitit')
    if notebook_name == 'auto_microbatching' and device == 'cpu':
        pytest.skip('auto_microbatching notebook only runs with a gpu')
    if notebook_name == 'TPU_Training_in_composer':
        pytest.skip('The CI does not support tpus')
    if notebook_name == 'ffcv_dataloaders' and device == 'cpu':
        pytest.skip('The FFCV notebook requires CUDA')
    if notebook_name == 'ffcv_dataloaders' and device == 'gpu':
        pytest.skip('CIFAR10 download is flaky')
    if notebook_name == 'finetune_huggingface':
        pytest.skip(
            "Error that is unreproducible locally: ModuleNotFoundError: No module named 'transformers.models.ernie_m.configuration_ernie_m'",
        )
    if notebook_name == 'pretrain_finetune_huggingface':
        pytest.skip(
            "Error that is unreproducible locally: No module named 'transformers.models.mega.configuration_mega'",
        )
    if notebook_name == 'checkpoint_autoresume':
        pytest.skip('MNIST dataset download is flaky')
    if notebook_name == 'exporting_for_inference':
        pytest.skip('MNIST dataset download is flaky')

    try:
        import boto3
    except ImportError as e:
        raise MissingConditionalImportError('streaming', 'boto3') from e

    obj = urlparse('s3://mosaicml-internal-integration-testing/read_only/CIFAR-10/')
    s3 = boto3.resource('s3')
    bucket = s3.Bucket(obj.netloc)  # pyright: ignore[reportGeneralTypeIssues]
    files = bucket.objects.filter(Prefix=obj.path.lstrip('/'))
    for file in files:
        target = os.path.join(os.getcwd(), 'data', os.path.relpath(file.key, obj.path.lstrip('/')))
        if not os.path.exists(target):
            os.makedirs(os.path.dirname(target), exist_ok=True)
        if file.key[-1] == '/':
            continue
        bucket.download_file(file.key, target)

    with testbook.testbook(notebook) as tb:
        tb.inject(trainer_monkeypatch_code)
        tb.inject('patch_notebooks()')
        for i, cell in enumerate(tb.cells):
            if cell['cell_type'] != 'code':
                continue
            cell['source'] = modify_cell_source(
                tb,
                notebook_name=notebook_name,
                cell_source=cell['source'],
                s3_bucket=s3_bucket,
            )
            tb.execute_cell(i)



# Copyright 2022 MosaicML Composer authors
# SPDX-License-Identifier: Apache-2.0

from typing import Mapping, Union

import pytest
import torch

from composer.core.data_spec import _split_list, _split_tensor, default_split_batch


def dummy_tensor_batch(batch_size=12) -> torch.Tensor:
    return torch.randn(size=(batch_size, 3, 32, 32))


def dummy_list_str(batch_size=12) -> list[str]:
    return [str(x) for x in range(batch_size)]


def dummy_tuple_batch(batch_size=12) -> list[torch.Tensor]:
    # pytorch default collate converts tuples to lists
    # https://github.com/pytorch/pytorch/blob/e451259a609acdcd83105177ddba73fc41cfa9b4/torch/utils/data/_utils/collate.py#L67
    image = torch.randn(size=(batch_size, 3, 32, 32))
    target = torch.randint(size=(batch_size,), high=10)
    return [image, target]


def dummy_tuple_batch_long(batch_size=12) -> list[torch.Tensor]:
    image_1 = torch.randn(size=(batch_size, 3, 32, 32))
    image_2 = torch.randn(size=(batch_size, 3, 32, 32))
    image_3 = torch.randn(size=(batch_size, 3, 32, 32))
    target = torch.randint(size=(batch_size,), high=10)
    return [image_1, image_2, image_3, target]


def dummy_dict_batch(batch_size=12) -> dict[str, torch.Tensor]:
    image = torch.randn(size=(batch_size, 3, 32, 32))
    target = torch.randint(size=(batch_size,), high=10)
    return {'image': image, 'target': target}


def dummy_dict_batch_with_metadata(batch_size=12) -> dict[str, Union[list, torch.Tensor, str]]:
    # sometimes metadata is included with a batch that isn't taken by the model.
    image = torch.randn(size=(batch_size, 3, 32, 32))
    target = torch.randint(size=(batch_size,), high=10)
    meta = ['hi im a tag' for _ in range(batch_size)]
    index = [[1, 2, 3] for _ in range(batch_size)]
    return {'image': image, 'target': target, 'meta': meta, 'index': index}


def dummy_dict_batch_with_common_metadata(batch_size=12) -> dict[str, Union[list, torch.Tensor, str]]:
    # sometimes metadata is included with a batch that isn't taken by the model.
    image = torch.randn(size=(batch_size, 3, 32, 32))
    target = torch.randint(size=(batch_size,), high=10)
    meta = 'this is a string'
    index = [[1, 2, 3] for _ in range(batch_size)]
    return {'image': image, 'target': target, 'meta': meta, 'index': index}


def dummy_maskrcnn_batch(
    batch_size=12,
    image_height=12,
    image_width=12,
    num_classes=80,
    max_detections=5,
) -> list[tuple[torch.Tensor, dict[str, torch.Tensor]]]:

    def generate_maskrcnn_sample(
        num_detections,
        image_height=image_height,
        image_width=image_width,
        num_classes=num_classes,
    ):
        """Generates a maskrcnn style sample: (Tensor, dict[Tensor])."""
        image = torch.randn(size=(3, image_height, image_width)).type(torch.float)
        target = {
            'boxes':
                torch.randint(size=(num_detections, 4), low=0, high=min(image_height, image_width)).type(torch.float),
            'labels':
                torch.randint(size=(num_detections,), low=0, high=num_classes + 1),
            'masks':
                torch.randint(size=(num_detections, image_height, image_width), low=0, high=2).type(torch.uint8),
        }
        return image, target

    return [
        generate_maskrcnn_sample(num_detections=n)
        for n in torch.randint(size=(batch_size,), low=1, high=max_detections + 1)
    ]


def dummy_batches(batch_size=12):
    return [
        dummy_tensor_batch(batch_size=batch_size),
        dummy_list_str(batch_size=batch_size),
        dummy_tuple_batch(batch_size=batch_size),
        dummy_tuple_batch_long(batch_size=batch_size),
        dummy_dict_batch(batch_size=batch_size),
        dummy_dict_batch_with_metadata(batch_size=batch_size),
        dummy_dict_batch_with_common_metadata(batch_size=batch_size),
    ]


@pytest.mark.parametrize('batch', dummy_batches(12))
def test_split_without_error(batch):
    microbatches = default_split_batch(batch, microbatch_size=3)
    assert len(microbatches) == 4


@pytest.mark.parametrize('batch', [dummy_tensor_batch(i) for i in [12, 13, 14, 15]])
def test_tensor_vs_list_chunking(batch):
    tensor_microbatches = _split_tensor(batch, microbatch_size=4)
    list_microbatches = _split_list(list(batch), microbatch_size=4)

    assert len(tensor_microbatches) == len(list_microbatches)
    assert all(torch.equal(t1, torch.stack(t2, dim=0)) for t1, t2 in zip(tensor_microbatches, list_microbatches))


@pytest.mark.parametrize('batch', [dummy_tuple_batch(12)])
def test_split_tuple(batch):
    microbatches = default_split_batch(batch, microbatch_size=4)
    # should be 3 microbatches of size 4 tensors pairs
    # should split into [(x, y), (x, y), (x, y)]
    assert len(microbatches[0]) == 2


@pytest.mark.parametrize('batch', [dummy_tuple_batch_long(12)])
def test_split_tuple_long(batch):
    microbatches = default_split_batch(batch, microbatch_size=4)
    assert len(microbatches[0]) == 4


@pytest.mark.parametrize('batch', dummy_batches(6))
def test_batch_sizes(batch):
    microbatches = default_split_batch(batch, microbatch_size=2)
    # should split into [len(2), len(2), len(1)]
    assert len(microbatches) == 3
    for microbatch in microbatches:
        if isinstance(microbatch, Mapping):
            assert len(microbatch['image']) == 2
            assert len(microbatch['target']) == 2
        if isinstance(microbatch, tuple):
            assert len(microbatch[0]) == 2
        if isinstance(microbatch, list):
            assert len(microbatch) == 2


@pytest.mark.parametrize('batch', dummy_batches(5))
def test_odd_batch_sizes(batch):
    microbatches = default_split_batch(batch, microbatch_size=2)
    # should split into [len(2), len(2), len(1)]
    assert len(microbatches) == 3
    last_microbatch = microbatches[-1]
    if isinstance(last_microbatch, Mapping):
        assert len(last_microbatch['image']) == 1
        assert len(last_microbatch['target']) == 1
    if isinstance(last_microbatch, tuple):
        assert len(last_microbatch[0]) == 1
    if isinstance(last_microbatch, list):
        assert len(last_microbatch) == 1


@pytest.mark.parametrize('batch', dummy_batches(2))
def test_microbatch_size_greater_than_batch_size(batch):
    with pytest.warns(UserWarning):
        microbatches = default_split_batch(batch, microbatch_size=3)
        assert len(microbatches) == 1


@pytest.mark.parametrize('batch', [dummy_maskrcnn_batch(12)])
def test_microbatch_size_split_maskrcnn(batch):
    microbatches = _split_list(batch, microbatch_size=4)
    assert len(microbatches) == 3


@pytest.mark.parametrize('batch', [dummy_dict_batch_with_common_metadata(12)])
def test_primitive_broadcast(batch):
    microbatches = default_split_batch(batch, microbatch_size=3)
    assert len(microbatches) == 4
    for mb in microbatches:
        assert mb['meta'] == 'this is a string'



# Copyright 2022 MosaicML Composer authors
# SPDX-License-Identifier: Apache-2.0

"""Object store base class and implementations."""

from composer.utils.object_store.gcs_object_store import GCSObjectStore
from composer.utils.object_store.libcloud_object_store import LibcloudObjectStore
from composer.utils.object_store.mlflow_object_store import (
    MLFLOW_EXPERIMENT_ID_FORMAT_KEY,
    MLFLOW_RUN_ID_FORMAT_KEY,
    MLFlowObjectStore,
)
from composer.utils.object_store.object_store import ObjectStore, ObjectStoreTransientError
from composer.utils.object_store.oci_object_store import OCIObjectStore
from composer.utils.object_store.s3_object_store import S3ObjectStore
from composer.utils.object_store.sftp_object_store import SFTPObjectStore
from composer.utils.object_store.uc_object_store import UCObjectStore

__all__ = [
    'ObjectStore',
    'ObjectStoreTransientError',
    'LibcloudObjectStore',
    'MLFlowObjectStore',
    'S3ObjectStore',
    'SFTPObjectStore',
    'OCIObjectStore',
    'GCSObjectStore',
    'UCObjectStore',
    'MLFLOW_EXPERIMENT_ID_FORMAT_KEY',
    'MLFLOW_RUN_ID_FORMAT_KEY',
]



# Copyright 2022 MosaicML Composer authors
# SPDX-License-Identifier: Apache-2.0

"""Periodically log generations from a set of prompts."""

import logging
import time
from typing import Any, Optional, Union, cast

from composer.core import Callback, Event, State, Time, get_precision_context
from composer.loggers import Logger
from composer.models import HuggingFaceModel
from composer.utils import create_interval_scheduler, dist
from composer.utils.import_helpers import MissingConditionalImportError

log = logging.getLogger(__name__)


class Generate(Callback):
    """Periodically log generations from a set of prompts.

    Args:
        prompts (list[str]): The list of prompts you would like to produce generations for
        interval (Union[str, int, :class:`.Time`]): The interval describing how often checkpoints should be
            saved. If an integer, it will be assumed to be in :attr:`.TimeUnit.EPOCH`.
            Otherwise, the unit must be either :attr:`.TimeUnit.EPOCH`, :attr:`.TimeUnit.BATCH`,
            :attr:`.TimeUnit.TOKEN`, or :attr:`.TimeUnit.SAMPLE`.
        batch_size (Optional[int]): Size of a prompt batch for generation. If None, defaults to the number of prompts.
        kwargs: All kwargs will be passed along to the call to generate. This is for things like `do_sample`, `top_p`, etc
    """

    def __init__(
        self,
        prompts: list[str],
        interval: Union[str, int, Time],
        batch_size: Optional[int] = None,
        **kwargs: Any,
    ):
        try:
            import transformers
        except ImportError as e:
            raise MissingConditionalImportError(
                extra_deps_group='nlp',
                conda_package='transformers',
                conda_channel='conda-forge',
            ) from e
        del transformers
        self.prompts = prompts
        self.generate_kwargs = kwargs
        self.batch_size = batch_size if batch_size is not None else len(prompts)
        self.check_interval = create_interval_scheduler(interval, include_end_of_training=True)
        self.last_generate_batch: Optional[Time] = None

    def run_event(self, event: Event, state: State, logger: Logger) -> None:
        if state.get_elapsed_duration(
        ) is not None and self.check_interval(state, event) and self.last_generate_batch != state.timestamp.batch:
            start = time.time()
            self.generate(state, logger)
            diff = time.time() - start
            log.info(f'Generate callback ran in {diff} seconds for {len(self.prompts)} prompts')

    def generate(self, state: State, logger: Logger):
        self.last_generate_batch = state.timestamp.batch

        model = state.model.module if state.is_model_ddp else state.model
        if not isinstance(model, HuggingFaceModel):  # TODO: Extend to support any models that have a generate method.
            raise ValueError(f'Expected HuggingFaceModel, but got {model.__class__.__name__}')

        if not hasattr(model, 'tokenizer') or model.tokenizer is None:
            raise ValueError(
                f'Model {model.__class__.__name__} does not have a tokenizer which is required for generation.',
            )
        tokenizer = model.tokenizer

        from transformers import PreTrainedTokenizerBase
        tokenizer = cast(PreTrainedTokenizerBase, tokenizer)

        # Set to evaluation mode and stash the original mode.
        original_mode = model.training
        model.eval()
        device = state.device

        # Stash the original value of padding_side because generation requires left padding
        original_padding_side = tokenizer.padding_side
        tokenizer.padding_side = 'left'
        if tokenizer.pad_token_id is None:
            tokenizer.pad_token_id = tokenizer.eos_token_id
        tokenized_input = tokenizer(self.prompts, return_tensors='pt', padding=True)

        all_input_ids = tokenized_input['input_ids']
        all_attn_masks = tokenized_input['attention_mask']

        output_token_ids = []
        # dummy forward call needed for FSDP to work consistently
        model.dummy_forward_called = False

        n_prompts = len(self.prompts)
        for start in range(0, n_prompts, self.batch_size):
            end = min(start + self.batch_size, n_prompts)
            input_ids = all_input_ids[start:end]  # pyright: ignore[reportGeneralTypeIssues]
            attn_mask = all_attn_masks[start:end]  # pyright: ignore[reportGeneralTypeIssues]

            # Move batch to device.
            input_ids = device.tensor_to_device(input_ids)
            attn_mask = device.tensor_to_device(attn_mask)
            with get_precision_context(state.precision):
                output_token_ids.extend(
                    model.generate(  # type: ignore
                        input_ids=input_ids,
                        attention_mask=attn_mask,
                        synced_gpus=dist.get_world_size() > 1,
                        **self.generate_kwargs,
                    ),
                )

        if dist.get_global_rank() == 0:
            # Process prompts and outputs into a table.
            rows = []
            input_tokens_len = all_input_ids.shape[1]  # pyright: ignore[reportGeneralTypeIssues]
            for i, prompt in enumerate(self.prompts):
                output_tokens = output_token_ids[i][input_tokens_len:]
                output_text = tokenizer.decode(output_tokens, skip_special_tokens=True)
                rows.append([prompt, output_text])

            logger.log_table(columns=['prompt', 'generation'], rows=rows, name='generations')

        tokenizer.padding_side = original_padding_side
        model.train(mode=original_mode)



# Copyright 2022 MosaicML Composer authors
# SPDX-License-Identifier: Apache-2.0

"""Core CutOut classes and functions."""

from __future__ import annotations

import logging
from typing import Any, Callable, Optional, TypeVar, Union

import numpy as np
import torch
from PIL.Image import Image as PillowImage
from torch import Tensor

from composer.algorithms.utils.augmentation_common import image_as_type
from composer.core import Algorithm, Event, State
from composer.loggers import Logger

log = logging.getLogger(__name__)

__all__ = ['CutOut', 'cutout_batch']

ImgT = TypeVar('ImgT', torch.Tensor, PillowImage)


def cutout_batch(input: ImgT, num_holes: int = 1, length: float = 0.5, uniform_sampling: bool = False) -> ImgT:
    """See :class:`.CutOut`.

    Args:
        input (PIL.Image.Image | torch.Tensor): Image or batch of images. If
            a :class:`torch.Tensor`, must be a single image of shape ``(C, H, W)``
            or a batch of images of shape ``(N, C, H, W)``.
        num_holes: Integer number of holes to cut out. Default: ``1``.
        length (float, optional): Relative side length of the masked region.
            If specified, ``length`` is interpreted as a fraction of ``H`` and
            ``W``, and the resulting box is a square with side length
            ``length * min(H, W)``. Must be in the interval :math:`(0, 1)`.
            Default: ``0.5``.
        uniform_sampling (bool, optional): If ``True``, sample the bounding
            box such that each pixel has an equal probability of being masked.
            If ``False``, defaults to the sampling used in the original paper
            implementation. Default: ``False``.

    Returns:
        X_cutout: Batch of images with ``num_holes`` square holes with
            dimension determined by ``length`` replaced with zeros.

    Example:
         .. testcode::

            from composer.algorithms.cutout import cutout_batch
            new_input_batch = cutout_batch(X_example, num_holes=1, length=0.25)
    """
    X_tensor = image_as_type(input, torch.Tensor)
    h = X_tensor.shape[-2]
    w = X_tensor.shape[-1]

    length = int(min(h, w) * length)

    mask = torch.ones_like(X_tensor)
    for _ in range(num_holes):
        if uniform_sampling is True:
            y = np.random.randint(-length // 2, high=h + length // 2)
            x = np.random.randint(-length // 2, high=w + length // 2)
        else:
            y = np.random.randint(h)
            x = np.random.randint(w)

        mask = _generate_mask(mask, w, h, x, y, length)

    X_cutout = X_tensor * mask
    X_out = image_as_type(X_cutout, input.__class__)  # pyright struggling with unions
    return X_out


class CutOut(Algorithm):
    """`CutOut <https://arxiv.org/abs/1708.04552>`_ is a data augmentation technique
    that works by masking out one or more square regions of an input image.

    This implementation cuts out the same square from all images in a batch.

    Example:
         .. testcode::

            from composer.algorithms import CutOut
            from composer.trainer import Trainer

            cutout_algorithm = CutOut(num_holes=1, length=0.25)
            trainer = Trainer(
                model=model,
                train_dataloader=train_dataloader,
                eval_dataloader=eval_dataloader,
                max_duration="1ep",
                algorithms=[cutout_algorithm],
                optimizers=[optimizer]
            )

    Args:
        num_holes (int, optional): Integer number of holes to cut out.
            Default: ``1``.
        length (float, optional): Relative side length of the masked region.
            If specified, ``length`` is interpreted as a fraction of ``H`` and
            ``W``, and the resulting box is a square with side length
            ``length * min(H, W)``. Must be in the interval :math:`(0, 1)`.
            Default: ``0.5``.
        uniform_sampling (bool, optional): If ``True``, sample the bounding
            box such that each pixel has an equal probability of being masked.
            If ``False``, defaults to the sampling used in the original paper
            implementation. Default: ``False``.
        input_key (str | int | tuple[Callable, Callable] | Any, optional): A key that indexes to the input
            from the batch. Can also be a pair of get and set functions, where the getter
            is assumed to be first in the pair.  The default is 0, which corresponds to any sequence, where the first element
            is the input. Default: ``0``.
    """

    def __init__(
        self,
        num_holes: int = 1,
        length: float = 0.5,
        uniform_sampling: bool = False,
        input_key: Union[str, int, Callable, Any] = 0,
    ):
        self.num_holes = num_holes
        self.length = length
        self.uniform_sampling = uniform_sampling
        self.input_key = input_key

    def match(self, event: Event, state: State) -> bool:
        return event == Event.AFTER_DATALOADER

    def apply(self, event: Event, state: State, logger: Logger) -> Optional[int]:
        x = state.batch_get_item(self.input_key)
        assert isinstance(x, Tensor), 'Multiple tensors not supported for Cutout.'

        new_x = cutout_batch(x, num_holes=self.num_holes, length=self.length, uniform_sampling=self.uniform_sampling)
        state.batch_set_item(self.input_key, new_x)


def _generate_mask(mask: Tensor, width: int, height: int, x: int, y: int, cutout_length: int) -> Tensor:
    y1 = np.clip(y - cutout_length // 2, 0, height)
    y2 = np.clip(y + cutout_length // 2, 0, height)
    x1 = np.clip(x - cutout_length // 2, 0, width)
    x2 = np.clip(x + cutout_length // 2, 0, width)

    mask[..., y1:y2, x1:x2] = 0.

    return mask



# Copyright 2022 MosaicML Composer authors
# SPDX-License-Identifier: Apache-2.0

"""Helper utilities."""

from composer.utils.auto_log_hparams import (
    convert_flat_dict_to_nested_dict,
    convert_nested_dict_to_flat_dict,
    extract_hparams,
)
from composer.utils.batch_helpers import batch_get, batch_set
from composer.utils.checkpoint import (
    PartialFilePath,
    get_save_filename,
    load_checkpoint,
    safe_torch_load,
    save_checkpoint,
)
from composer.utils.collect_env import (
    configure_excepthook,
    disable_env_report,
    enable_env_report,
    get_composer_env_dict,
    print_env,
)
from composer.utils.compression import (
    KNOWN_COMPRESSORS,
    CliCompressor,
    get_compressor,
    is_compressed_pt,
)
from composer.utils.device import get_device, is_hpu_installed, is_xla_installed
from composer.utils.eval_client import EvalClient, LambdaEvalClient, LocalEvalClient, MosaicMLLambdaEvalClient
from composer.utils.file_helpers import (
    FORMAT_NAME_WITH_DIST_AND_TIME_TABLE,
    FORMAT_NAME_WITH_DIST_TABLE,
    create_symlink_file,
    ensure_folder_has_no_conflicting_files,
    ensure_folder_is_empty,
    format_name_with_dist,
    format_name_with_dist_and_time,
    get_file,
    is_tar,
    maybe_create_object_store_from_uri,
    maybe_create_remote_uploader_downloader_from_uri,
    parse_uri,
)
from composer.utils.import_helpers import MissingConditionalImportError, import_object
from composer.utils.inference import ExportFormat, Transform, export_for_inference, export_with_logger, quantize_dynamic
from composer.utils.iter_helpers import IteratorFileStream, ensure_tuple, map_collection
from composer.utils.misc import (
    STR_TO_DTYPE,
    ParallelismType,
    add_vision_dataset_transform,
    create_interval_scheduler,
    get_free_tcp_port,
    is_model_deepspeed,
    is_model_fsdp,
    is_notebook,
    model_eval_mode,
    partial_format,
)
from composer.utils.object_store import (
    MLFLOW_EXPERIMENT_ID_FORMAT_KEY,
    MLFLOW_RUN_ID_FORMAT_KEY,
    GCSObjectStore,
    LibcloudObjectStore,
    MLFlowObjectStore,
    ObjectStore,
    ObjectStoreTransientError,
    OCIObjectStore,
    S3ObjectStore,
    SFTPObjectStore,
    UCObjectStore,
)
from composer.utils.parallelism import FSDPConfig, ParallelismConfig, TPConfig, create_fsdp_config
from composer.utils.retrying import retry
from composer.utils.string_enum import StringEnum
from composer.utils.warnings import VersionedDeprecationWarning

__all__ = [
    'ensure_tuple',
    'get_free_tcp_port',
    'map_collection',
    'IteratorFileStream',
    'FORMAT_NAME_WITH_DIST_AND_TIME_TABLE',
    'FORMAT_NAME_WITH_DIST_TABLE',
    'get_file',
    'PartialFilePath',
    'create_symlink_file',
    'ObjectStore',
    'ObjectStoreTransientError',
    'LibcloudObjectStore',
    'S3ObjectStore',
    'SFTPObjectStore',
    'MLFlowObjectStore',
    'OCIObjectStore',
    'GCSObjectStore',
    'UCObjectStore',
    'MissingConditionalImportError',
    'get_save_filename',
    'import_object',
    'is_model_deepspeed',
    'is_model_fsdp',
    'is_notebook',
    'StringEnum',
    'load_checkpoint',
    'save_checkpoint',
    'safe_torch_load',
    'ensure_folder_is_empty',
    'ensure_folder_has_no_conflicting_files',
    'export_for_inference',
    'export_with_logger',
    'quantize_dynamic',
    'format_name_with_dist',
    'format_name_with_dist_and_time',
    'is_tar',
    'maybe_create_object_store_from_uri',
    'maybe_create_remote_uploader_downloader_from_uri',
    'parse_uri',
    'batch_get',
    'batch_set',
    'configure_excepthook',
    'disable_env_report',
    'enable_env_report',
    'print_env',
    'get_composer_env_dict',
    'retry',
    'model_eval_mode',
    'get_device',
    'is_xla_installed',
    'is_hpu_installed',
    'ExportFormat',
    'Transform',
    'export_with_logger',
    'extract_hparams',
    'convert_nested_dict_to_flat_dict',
    'convert_flat_dict_to_nested_dict',
    'create_interval_scheduler',
    'EvalClient',
    'LambdaEvalClient',
    'LocalEvalClient',
    'MosaicMLLambdaEvalClient',
    'partial_format',
    'add_vision_dataset_transform',
    'VersionedDeprecationWarning',
    'is_compressed_pt',
    'CliCompressor',
    'get_compressor',
    'KNOWN_COMPRESSORS',
    'STR_TO_DTYPE',
    'ParallelismType',
    'create_fsdp_config',
    'FSDPConfig',
    'TPConfig',
    'ParallelismConfig',
    'MLFLOW_EXPERIMENT_ID_FORMAT_KEY',
    'MLFLOW_RUN_ID_FORMAT_KEY',
]



# Copyright 2022 MosaicML Composer authors
# SPDX-License-Identifier: Apache-2.0

import pytest
from torch.utils.data import DataLoader

from composer.core import DataSpec
from composer.trainer import Trainer
from composer.utils import dist, get_device
from tests.common import device
from tests.common.datasets import RandomTextClassificationDataset, RandomTextLMDataset
from tests.common.models import SimpleTransformerClassifier, SimpleTransformerMaskedLM


def test_simple_nlp_classification():
    vocab_size = 100
    sequence_length = 32
    num_classes = 2
    size = 96
    batch_size = 8

    train_dataset = RandomTextClassificationDataset(
        size=size,
        vocab_size=vocab_size,
        sequence_length=sequence_length,
        num_classes=num_classes,
    )
    eval_dataset = RandomTextClassificationDataset(
        size=size,
        vocab_size=vocab_size,
        sequence_length=sequence_length,
        num_classes=num_classes,
    )
    predict_dataset = RandomTextClassificationDataset(
        size=size,
        vocab_size=vocab_size,
        sequence_length=sequence_length,
        num_classes=num_classes,
    )

    model = SimpleTransformerClassifier(vocab_size=vocab_size, num_classes=num_classes)

    train_dataloader = DataLoader(train_dataset, batch_size=batch_size, sampler=dist.get_sampler(train_dataset))
    eval_dataloader = DataLoader(eval_dataset, batch_size=batch_size, sampler=dist.get_sampler(eval_dataset))
    predict_dataloader = DataLoader(predict_dataset, batch_size=8)

    trainer = Trainer(
        model=model,
        train_dataloader=train_dataloader,
        max_duration='2ep',
        eval_dataloader=eval_dataloader,
    )

    trainer.fit()
    trainer.eval()

    # Check that there is some train/eval accuracy
    assert trainer.state.train_metrics is not None
    assert trainer.state.train_metrics['MulticlassAccuracy'].compute() != 0.0
    assert trainer.state.eval_metrics['eval']['MulticlassAccuracy'].compute() != 0.0

    predictions = trainer.predict(predict_dataloader)

    # Check that the output predictions are the expected shape
    num_predict_batches_expected = ((size - 1) // batch_size) + 1
    assert len(predictions) == num_predict_batches_expected
    assert predictions[0].shape == (batch_size, 2)


def test_simple_nlp_mlm(tiny_bert_tokenizer, tiny_bert_model):
    transformers = pytest.importorskip('transformers')

    vocab_size = tiny_bert_tokenizer.vocab_size
    sequence_length = 32
    size = 96
    batch_size = 8

    train_dataset = RandomTextLMDataset(
        size=size,
        vocab_size=vocab_size,
        sequence_length=sequence_length,
        use_keys=True,
    )
    eval_dataset = RandomTextLMDataset(size=size, vocab_size=vocab_size, sequence_length=sequence_length, use_keys=True)
    predict_dataset = RandomTextLMDataset(
        size=size,
        vocab_size=vocab_size,
        sequence_length=sequence_length,
        use_keys=True,
    )
    collator = transformers.DataCollatorForLanguageModeling(tokenizer=tiny_bert_tokenizer, mlm_probability=0.15)

    model = SimpleTransformerMaskedLM(vocab_size=vocab_size)

    train_dataloader = DataLoader(
        train_dataset,
        batch_size=batch_size,
        sampler=dist.get_sampler(train_dataset),
        collate_fn=collator,
    )
    eval_dataloader = DataLoader(
        eval_dataset,
        batch_size=batch_size,
        sampler=dist.get_sampler(eval_dataset),
        collate_fn=collator,
    )
    predict_dataloader = DataLoader(predict_dataset, batch_size=8)

    trainer = Trainer(
        model=model,
        train_dataloader=train_dataloader,
        max_duration='2ep',
        eval_dataloader=eval_dataloader,
    )

    trainer.fit()
    trainer.eval()

    # Check that there is some train/eval cross entropy
    assert trainer.state.train_metrics is not None
    assert trainer.state.train_metrics['LanguageCrossEntropy'].compute() != 0.0
    assert trainer.state.eval_metrics['eval']['LanguageCrossEntropy'].compute() != 0.0

    predictions = trainer.predict(predict_dataloader)

    # Check that the output predictions are the expected shape
    num_predict_batches_expected = ((size - 1) // batch_size) + 1
    assert len(predictions) == num_predict_batches_expected
    assert predictions[0].shape == (batch_size, sequence_length, vocab_size)


@device('gpu')
def test_simple_nlp_mlm_token_batch(tiny_bert_tokenizer, device):
    transformers = pytest.importorskip('transformers')

    vocab_size = tiny_bert_tokenizer.vocab_size
    sequence_length = 32
    size = 96
    batch_size = 8
    device = get_device(device)

    train_dataset = RandomTextLMDataset(
        size=size,
        vocab_size=vocab_size,
        sequence_length=sequence_length,
        use_keys=True,
        pad_token_id=tiny_bert_tokenizer.pad_token_id,
    )
    for i in range(size):  # Proactively load dataset for consistent randomization
        train_dataset[i]
    collator = transformers.DataCollatorForLanguageModeling(tokenizer=tiny_bert_tokenizer)

    # Get the model's state dict before training starts, so we can reproduce results
    model = SimpleTransformerMaskedLM(vocab_size=vocab_size)
    state_dict = model.state_dict()

    # Set up the data spec that can count the non-padding tokens in a batch
    train_dataloader = DataLoader(
        train_dataset,
        batch_size=batch_size,
        sampler=dist.get_sampler(train_dataset),
        collate_fn=collator,
    )
    data_spec = DataSpec(
        dataloader=train_dataloader,
        get_num_tokens_in_batch=lambda b: (b['input_ids'] != tiny_bert_tokenizer.pad_token_id).sum().item(),
    )

    trainer = Trainer(
        model=model,
        seed=42,
        train_dataloader=data_spec,
        max_duration='2ep',
        device_train_microbatch_size=batch_size // 2,
        accumulate_train_batch_on_tokens=False,
        device=device,
    )
    trainer.fit()

    # Check that there is some train cross entropy
    assert trainer.state.train_metrics is not None
    cross_entropy = trainer.state.train_metrics['LanguageCrossEntropy'].compute()
    assert cross_entropy != 0.0

    # Set up a trainer that accumulates train loss based on token counts, after reloading original state dict
    model.load_state_dict(state_dict)
    token_trainer = Trainer(
        model=model,
        seed=42,
        train_dataloader=data_spec,
        max_duration='2ep',
        device_train_microbatch_size=batch_size // 2,
        accumulate_train_batch_on_tokens=True,
        device=device,
    )
    token_trainer.fit()

    # Check that there is some train cross entropy
    assert token_trainer.state.train_metrics is not None
    token_cross_entropy = token_trainer.state.train_metrics['LanguageCrossEntropy'].compute()
    assert token_cross_entropy != 0.0

    # Require that the train cross entropies are different between the trainers
    assert cross_entropy != token_cross_entropy

    # Make sure we can reproduce the original cross entropy calculation
    model.load_state_dict(state_dict)
    trainer2 = Trainer(
        model=model,
        seed=42,
        train_dataloader=data_spec,
        max_duration='2ep',
        device_train_microbatch_size=batch_size // 2,
        accumulate_train_batch_on_tokens=False,
        device=device,
    )
    trainer2.fit()
    assert trainer2.state.train_metrics is not None
    assert trainer2.state.train_metrics['LanguageCrossEntropy'].compute() == cross_entropy



# Copyright 2022 MosaicML Composer authors
# SPDX-License-Identifier: Apache-2.0

"""Retry helper."""

from __future__ import annotations

import collections.abc
import functools
import inspect
import logging
import random
import time
from typing import Any, Callable, Sequence, Type, TypeVar, Union, cast, overload

TCallable = TypeVar('TCallable', bound=Callable)

__all__ = ['retry']

log = logging.getLogger(__name__)


@overload
def retry(
    exc_class: Union[Type[Exception], Sequence[Type[Exception]]] = ...,
    num_attempts: int = ...,
    initial_backoff: float = ...,
    max_jitter: float = ...,
) -> Callable[[TCallable], TCallable]:
    ...


@overload
def retry(exc_class: TCallable) -> TCallable:
    # Use the decorator without parenthesis
    ...


# error: Type "(TCallable@retry) -> TCallable@retry" cannot be assigned to type "(func: Never) -> Never"
def retry(  # type: ignore
    exc_class: Union[TCallable, Type[Exception], Sequence[Type[Exception]]] = Exception,
    num_attempts: int = 3,
    initial_backoff: float = 1.0,
    max_jitter: float = 0.5,
):
    """Decorator to retry a function with backoff and jitter.

    Attempts are spaced out with ``initial_backoff + 2**num_attempts + random.random() * max_jitter`` seconds.

    Optionally, the decorated function can specify `retry_index` as an argument to receive the current attempt number.

    Example:
    .. testcode::

        from composer.utils import retry

        @retry(RuntimeError, num_attempts=3, initial_backoff=0.1)
        def flaky_function(retry_index: int):
            if retry_index < 2:
                raise RuntimeError("Called too soon!")
            return "Third time's a charm."

        print(flaky_function())

    .. testoutput::

        Third time's a charm.

    Args:
        exc_class (Type[Exception] | Sequence[Type[Exception]]], optional): The exception class or classes to retry.
            Defaults to Exception.
        num_attempts (int, optional): The total number of attempts to make. Defaults to 3.
        initial_backoff (float, optional): The initial backoff, in seconds. Defaults to 1.0.
        max_jitter (float, optional): The maximum amount of random jitter to add. Defaults to 0.5.

            Increasing the ``max_jitter`` can help prevent overloading a resource when multiple processes in parallel
            are calling the same underlying function.
    """
    if num_attempts < 1:
        raise ValueError('num_attempts must be at leats 1')

    def wrapped_func(func: TCallable) -> TCallable:

        @functools.wraps(func)
        def new_func(*args: Any, **kwargs: Any):
            retry_index_param = 'retry_index'
            i = 0
            while True:
                try:
                    if retry_index_param in inspect.signature(func).parameters:
                        kwargs[retry_index_param] = i
                    return func(*args, **kwargs)
                except exc_class as e:
                    log.debug(f'Attempt {i} failed. Exception type: {type(e)}, message: {str(e)}.')
                    if i + 1 == num_attempts:
                        raise e
                    else:
                        time.sleep(initial_backoff * 2**i + random.random() * max_jitter)
                        i += 1

        return cast(TCallable, new_func)

    if not isinstance(
        exc_class,
        collections.abc.Sequence,
    ) and not (isinstance(exc_class, type) and issubclass(exc_class, Exception)):
        # Using the decorator without (), like @retry_with_backoff
        func = cast(TCallable, exc_class)
        exc_class = Exception

        return wrapped_func(func)

    return wrapped_func



# Copyright 2024 MosaicML Composer authors
# SPDX-License-Identifier: Apache-2.0

"""Useful functions for generating state dicts and manipulating them."""

import fnmatch
import logging
import sys
from typing import TYPE_CHECKING, Any, Dict, Iterable, List, Optional, Sequence, Union

from torch.utils.data import DataLoader, Dataset

from composer.core.data_spec import DataSpec

if TYPE_CHECKING:
    from composer.core.evaluator import Evaluator

import torch
from packaging import version
from torch import nn
from torch.distributed.fsdp import FullyShardedDataParallel as FSDP
from torch.nn.parallel import DistributedDataParallel
from torch.utils.data import DataLoader

from composer.core.evaluator import Evaluator
from composer.core.state import State
from composer.core.time import Timestamp
from composer.devices import Device
from composer.models import ComposerModel, HuggingFaceModel
from composer.utils import STR_TO_DTYPE, dist, get_composer_env_dict, reproducibility

log = logging.getLogger(__name__)

__all__ = ['get_model_state_dict', 'get_optim_state_dict', 'get_metadata_state_dict', 'get_resumption_state_dict']


def get_model_state_dict(
    model: Union[ComposerModel, nn.Module],
    sharded_state_dict: bool = False,
    precision: Union[str, torch.dtype] = 'fp32',
    include_keys: Optional[Union[str, Sequence[str]]] = None,
    ignore_keys: Optional[Union[str, Sequence[str]]] = None,
    cpu_offload: Optional[bool] = None,
) -> dict[str, Any]:
    """Generate the state dict of the model.

    Args:
        model: The model to get the state dict from.
        sharded_state_dict: Whether the model state dict should be sharded or not. If True, every rank returns the state dict of its shards.
            If False, then rank 0 returns the state dict of the entire model and the other ranks return an empty dict. Default is False.
        precision: The precision of the model. Can be specified as a string ('fp32', 'fp16', 'bf16') or a torch.dtype.
        include_keys: The list of keys to exclusively include in the state dict. If None, all keys are included. Both include_keys and ignore_keys cannot be non-None.
        ignore_keys: The list of keys to ignore in the state dict. If None, no keys are ignored. Both include_keys and ignore_keys cannot be non-None.
        cpu_offload: Whether to offload the state dict to CPU. If None, it is set to True if FSDP is enabled, False otherwise.

    Returns:
        The state dict of the model.
    """
    if include_keys is not None and ignore_keys is not None:
        raise ValueError(f'Both {include_keys=} and {ignore_keys=} cannot be non-None.')

    is_fsdp = _is_model_fsdp(model)
    if not is_fsdp and sharded_state_dict:
        raise ValueError('Sharded state dict can only be generated for FSDP models.')
    cpu_offload = cpu_offload if cpu_offload is not None else (is_fsdp and not sharded_state_dict)

    log.debug('Extracting model state dict')
    if version.parse(torch.__version__) >= version.parse('2.2.0') and dist.is_initialized():
        from torch.distributed.checkpoint import state_dict as DCPSD  # Distributed Checkpoint State Dict
        from torch.distributed.checkpoint.state_dict import StateDictOptions

        use_unsharded_state_dict = not sharded_state_dict

        log.debug('Calling torch get_model_state_dict...')
        model_state_dict = DCPSD.get_model_state_dict(
            model=model,
            submodules=None,  # We extract submodules below
            options=StateDictOptions(
                full_state_dict=use_unsharded_state_dict,
                cpu_offload=cpu_offload,
            ),
        )
    else:
        if is_fsdp:
            log.debug('Calling legacy FSDP context manager to get model state dict...')
            model_state_dict = _get_model_state_dict_with_fsdp_context_manager(model, sharded_state_dict, cpu_offload)
        else:
            log.debug('Calling model.state_dict() for non-FSDP model...')
            model_state_dict = model.state_dict()
        if isinstance(model, DistributedDataParallel):
            nn.modules.utils.consume_prefix_in_state_dict_if_present(model_state_dict, 'module.')

    if include_keys is not None:
        model_state_dict = _extract_keys_from_state_dict(model_state_dict, include_keys)

    if ignore_keys is not None:
        model_state_dict = _remove_keys_from_state_dict(model_state_dict, ignore_keys)

    model_state_dict = _cast_state_dict_to_precision(state_dict=model_state_dict, precision=precision)

    log.debug('Finished extracting model state dict')
    return model_state_dict


def _cast_state_dict_to_precision(state_dict: Dict[str, Any], precision: Union[str, torch.dtype]) -> Dict[str, Any]:
    if isinstance(precision, str):
        precision = STR_TO_DTYPE[precision]

    new_state_dict = {k: v.to(precision) for k, v in state_dict.items()}
    return new_state_dict


def _extract_keys_from_state_dict(state_dict: dict[str, Any], include_keys: Union[str, Sequence[str]]):
    if isinstance(include_keys, str):
        include_keys = [include_keys]
    new_state_dict = {k: v for k, v in state_dict.items() if any(fnmatch.fnmatch(k, key) for key in include_keys)}

    return new_state_dict


def _remove_keys_from_state_dict(state_dict: dict[str, Any], ignore_keys: Union[str, Sequence[str]]):
    if isinstance(ignore_keys, str):
        ignore_keys = [ignore_keys]
    new_state_dict = {k: v for k, v in state_dict.items() if not any(fnmatch.fnmatch(k, key) for key in ignore_keys)}
    return new_state_dict


def _is_model_fsdp(model) -> bool:
    """Indicates if FSDP is enabled.

    Args:
        model: The model to check if FSDP is enabled.

    Returns:
        True if FSDP is enabled, False otherwise.

    """
    for module in model.modules():
        if isinstance(module, FSDP):
            return True
    return False


def _get_model_state_dict_with_fsdp_context_manager(model: nn.Module, sharded_state_dict: bool,
                                                    cpu_offload: bool) -> dict[str, Any]:
    """Get the model state dict with the FSDP context manager.

    Args:
        model: The model to get the state dict from.
        sharded: Whether the model state dict should be sharded or not. If True, every rank returns the state dict of its shards.
            If False, then rank 0 returns the state dict of the entire model.

    Returns:
        The state dict of the model.
    """
    from torch.distributed.fsdp.fully_sharded_data_parallel import (
        FullStateDictConfig,
        ShardedStateDictConfig,
        StateDictType,
    )
    state_dict_type = StateDictType.SHARDED_STATE_DICT if sharded_state_dict else StateDictType.FULL_STATE_DICT
    state_dict_config = ShardedStateDictConfig(offload_to_cpu=cpu_offload,
                                              ) if sharded_state_dict else FullStateDictConfig(
                                                  rank0_only=True,
                                                  offload_to_cpu=cpu_offload,
                                              )
    with FSDP.state_dict_type(model, state_dict_type=state_dict_type, state_dict_config=state_dict_config):
        model_state_dict = model.state_dict()
    return model_state_dict


def get_resumption_state_dict(state: State) -> Dict[str, Any]:
    """Generate the state dict for any objects needed for resumption.

    This includes:
        * timestamp
        * scheduler
        * dataset_state
        * scaler
        * rank_zero_seed
        * callbacks
        * algorithms

    Returns:
        The state dict containing the objects needed for resumption.
    """
    resumption_state_dict = {}
    resumption_state_dict['dataset_state'] = get_dataset_state_dict(
        state.train_dataloader,
        state.timestamp,
    )
    resumption_state_dict['timestamp'] = state.timestamp.state_dict()

    scheduler_state_dict = _make_state_dict_for_list_of_objects(state.schedulers)
    if scheduler_state_dict != {}:
        resumption_state_dict['schedulers'] = scheduler_state_dict

    # Use list of tuples to account for duplicates
    callbacks_state_dict = _make_state_dict_for_list_of_objects(state.callbacks, use_list_of_tuples=True)
    if callbacks_state_dict != {}:
        resumption_state_dict['callbacks'] = callbacks_state_dict

    # Use list of tuples to preserve order.
    algorithms_state_dict = _make_state_dict_for_list_of_objects(state.algorithms, use_list_of_tuples=True)
    if algorithms_state_dict != {}:
        resumption_state_dict['algorithms'] = algorithms_state_dict

    if state.scaler is not None:
        scaler_sd = _make_state_dict_for_list_of_objects(state.scaler)
        if scaler_sd != {}:
            resumption_state_dict['scaler'] = state.scaler.state_dict()

    resumption_state_dict['rank_zero_seed'] = state.rank_zero_seed
    resumption_state_dict['run_name'] = state.run_name
    resumption_state_dict['rng'] = reproducibility.get_rng_state()

    return resumption_state_dict


def _make_state_dict_for_list_of_objects(objects: Union[Sequence[Any], Any],
                                         use_list_of_tuples=False) -> Union[Dict[str, Any], List]:
    object_list = []
    object_dict = {}
    if not isinstance(objects, Sequence):
        objects = [objects]
    for obj in objects:
        if not hasattr(obj, 'state_dict') or obj.state_dict() == {}:
            continue
        if use_list_of_tuples:
            object_list.append((type(obj).__qualname__, obj.state_dict()))
        else:
            object_dict[type(obj).__qualname__] = obj.state_dict()
    if use_list_of_tuples:
        return object_list
    else:
        return object_dict


def get_dataset_state_dict(
    train_dataloader: Optional[Union[DataLoader, Iterable]],
    timestamp: Timestamp,
) -> Dict[str, Any]:
    """Collect the state dict(s) of our train and eval dataset(s).

    Returns:
        Dict[str, Any]: The state dict(s).
    """
    dataset_state_dict = {
        'train': None,
    }
    dataset = _dataset_of(train_dataloader)
    if hasattr(dataset, 'state_dict'):
        num_samples = int(timestamp.sample_in_epoch.value)
        dataset_state_dict['train'] = dataset.state_dict(num_samples, True)  # pyright: ignore

    return dataset_state_dict


def _get_optim_state_dict_with_fsdp_context_manager(
    model: nn.Module,
    optimizer: torch.optim.Optimizer,
    sharded_state_dict: bool,
    cpu_offload: bool,
) -> Dict[str, Any]:
    """Get the optimizer state dict with the FSDP context manager.

    Args:
        model: The model containing the parameters that the optimizer is optimizing.
        optimizer: The optimizer to get the state dict from.
        sharded_state_dict: Whether the optimizer state dict should be sharded or not. If True, every rank returns the state dict of its shards.
            If False, then rank 0 returns the state dict of the entire optimizer.
        cpu_offload: Whether to offload the state dict to CPU.

    Returns:
        The state dict of the optimizer.

    """
    from torch.distributed.fsdp.fully_sharded_data_parallel import (
        FullOptimStateDictConfig,
        FullStateDictConfig,
        ShardedOptimStateDictConfig,
        ShardedStateDictConfig,
        StateDictType,
    )
    state_dict_type = StateDictType.SHARDED_STATE_DICT if sharded_state_dict else StateDictType.FULL_STATE_DICT

    state_dict_config = ShardedStateDictConfig(offload_to_cpu=cpu_offload,
                                              ) if sharded_state_dict else FullStateDictConfig(
                                                  rank0_only=True,
                                                  offload_to_cpu=cpu_offload,
                                              )
    optim_state_dict_config = ShardedOptimStateDictConfig(
        offload_to_cpu=cpu_offload,
    ) if sharded_state_dict else FullOptimStateDictConfig(rank0_only=True, offload_to_cpu=cpu_offload)
    with FSDP.state_dict_type(
        model,
        state_dict_type=state_dict_type,
        state_dict_config=state_dict_config,
        optim_state_dict_config=optim_state_dict_config,
    ):
        optim_state_dict = FSDP.optim_state_dict(model, optimizer)
    return optim_state_dict


def get_optim_state_dict(
    model: Union[ComposerModel, nn.Module],
    optimizer: torch.optim.Optimizer,
    sharded_state_dict: bool = False,
    precision: str = 'fp32',
    include_keys: Optional[Union[str, Sequence[str]]] = None,
    ignore_keys: Optional[Union[str, Sequence[str]]] = None,
    cpu_offload: Optional[bool] = None,
) -> Dict[str, Any]:
    """Generate the state dict of the optimizer.

    Args:
        model: The model containing the parameters that the optimizer is optimizing.
        optimizer: The optimizer to get the state dict from.
        sharded: Whether the optimizer is sharded or not. If True, every rank returns the state dict of its shards.
            If False, then rank 0 returns the state dict of the entire optimizer and all other ranks return an empty dict.
        precision: The precision of the optimizer.
        include_keys: The list of keys to exclusively include in the state dict. If None, all keys are included. Both include_keys and ignore_keys cannot be non-None.
        ignore_keys: The list of keys to ignore in the state dict. If None, no keys are ignored. Both include_keys and ignore_keys cannot be non-None.
        cpu_offload: Whether to offload the state dict to CPU. If None, it is set to True if FSDP is enabled with non-sharded state dict and False otherwise.

    Returns:
        The state dict of the optimizer.
    """
    if include_keys is not None and ignore_keys is not None:
        raise ValueError(f'Both {include_keys=} and {ignore_keys=} cannot be non-None.')

    is_fsdp = _is_model_fsdp(model)
    if not is_fsdp and sharded_state_dict:
        raise ValueError('Sharded optim state dict can only be generated for FSDP models.')

    cpu_offload = cpu_offload if cpu_offload is not None else (is_fsdp and not sharded_state_dict)
    log.debug('Extracting optim state dict')
    if version.parse(torch.__version__) >= version.parse('2.2.0') and dist.is_initialized():
        from torch.distributed.checkpoint.state_dict import StateDictOptions, get_optimizer_state_dict
        log.debug('Calling torch get_optimizer_state_dict...')
        optim_state_dict: Dict[str, Any] = get_optimizer_state_dict(
                model=model,
                optimizers=optimizer,
                submodules=None, # We extract submodules below
                options=StateDictOptions(
                    full_state_dict=not sharded_state_dict,
                    cpu_offload=cpu_offload,
                ),
            )
    else:
        if is_fsdp:
            log.debug('Calling legacy FSDP context manager to get optim state dict...')
            optim_state_dict = _get_optim_state_dict_with_fsdp_context_manager(
                model,
                optimizer,
                sharded_state_dict,
                cpu_offload,
            )
        else:
            optim_state_dict = optimizer.state_dict()

    # For sharded models with non-sharded state dicts, only rank 0 has the full state dict including all the keys
    target_state_dict_on_this_rank = (not sharded_state_dict and dist.get_global_rank() == 0) or sharded_state_dict

    if target_state_dict_on_this_rank:
        if ignore_keys is not None:
            raise NotImplementedError('Ignoring keys in the optimizer state dict is not supported yet.')
        if include_keys is not None:
            raise NotImplementedError('Ignoring keys in the optimizer state dict is not supported yet.')

        # param_key := index (0,1,2,..., len(model.parameters())-1) for unsharded models.
        # param_key := fqn for sharded models.
        for param_key, param_state_dict in optim_state_dict['state'].items():
            optim_state_dict['state'][param_key] = _cast_state_dict_to_precision(param_state_dict, precision)
    return optim_state_dict


def get_metadata_state_dict(
    model: Optional[Union[ComposerModel, nn.Module]] = None,
    sharded_state_dict: Optional[bool] = None,
    precision: Optional[Union[str, torch.dtype]] = None,
    device: Optional[Device] = None,
    device_train_microbatch_size: Optional[int] = None,
) -> dict[str, Any]:
    """Generate the metadata and integrations for a training run.

    Args:
        model: The model to get the metadata state dict from. Only applicable if model is HuggingFaceModel
        sharded_state_dict: Whether the checkpoint this metadata state dict is associated with is sharded or not.
            Optional argument because this function may not be called in the context of making a full checkpoint.
        precision: The precision of the model. Can be specified as a string ('fp32', 'fp16', 'bf16') or a torch.dtype.
        device: The device the model is on.
        device_train_microbatch_size: The microbatch size used for training on the device.

    This state dict includes:
        * composer version
        * composer commit hash
        * gpu model
        * num nodes
        * num gpus
        * num gpus per node
        * cpu core count
        * cpu model
        * torch version
        * python version
        * optionally
            * dist/communication backend
            * precision
            * hf model metadata
            * device_train_microbatch_size
            * sharded vs unsharded state dict
            * model name
            * param fqns, shapes, and requires_grad
            * huggingface metadata

    Returns:
        The state dict containing the metadata and any integrations for a training run.
    """
    metadata_state_dict = get_composer_env_dict()

    python_version = '.'.join([str(getattr(sys.version_info, k)) for k in ['major', 'minor', 'micro']])

    metadata_state_dict['torch_version'] = torch.__version__
    metadata_state_dict['python_version'] = python_version
    if sharded_state_dict is not None:
        metadata_state_dict['sharded_state_dict'] = sharded_state_dict

    if model is not None:
        if isinstance(model, HuggingFaceModel):
            metadata_state_dict['huggingface'] = model.get_metadata()
            metadata_state_dict['model_name'] = model.model.__class__.__name__
        elif (isinstance(model, DistributedDataParallel) or
              isinstance(model, FSDP)) and isinstance(model.module, HuggingFaceModel):
            metadata_state_dict['huggingface'] = model.module.get_metadata()
            metadata_state_dict['model_name'] = model.module.model.__class__.__name__
        elif isinstance(model, FSDP) or isinstance(model, DistributedDataParallel):
            metadata_state_dict['model_name'] = model.module.__class__.__name__
        else:
            metadata_state_dict['model_name'] = model.__class__.__name__

    if device is not None:
        metadata_state_dict['dist_backend'] = device.dist_backend

    if device_train_microbatch_size:
        metadata_state_dict['device_train_microbatch_size'] = device_train_microbatch_size

    if precision is not None:
        if isinstance(precision, str):
            metadata_state_dict['precision'] = precision
        else:
            dtype_to_str = {v: k for k, v in STR_TO_DTYPE.items()}
            metadata_state_dict['precision'] = dtype_to_str[precision]
    else:
        metadata_state_dict['precision'] = 'fp32'

    return metadata_state_dict


def _dataset_of(dataloader: Optional[Union[Evaluator, DataSpec, DataLoader, Iterable]]) -> Optional[Dataset]:
    """Get the dataset contained by the given dataloader-like object.

    Args:
        dataloader (Evaluator | DataSpec | DataLoader | Iterable, optional): The dataloader, wrapped dataloader, or
            generic python iterable to get the dataset of, if applicable.

    Returns:
        Dataset: Its dataset, if there is one.
    """
    from composer.core.evaluator import Evaluator

    # If it's None, no dataset for you.
    if dataloader is None:
        return None

    # An Evaluator is a dataloader wrapped with metrics. Unwrap its dataloader.
    if isinstance(dataloader, Evaluator):
        dataloader = dataloader.dataloader

    # A DataSpec is a dataloader wrapped with an on-device transform. Unwrap its dataloader.
    if isinstance(dataloader, DataSpec):
        dataloader = dataloader.dataloader

    # If what we now have is an actual DataLoader, return its dataset. If not, return None.
    if isinstance(dataloader, DataLoader):
        return dataloader.dataset
    else:
        return None



# Copyright 2022 MosaicML Composer authors
# SPDX-License-Identifier: Apache-2.0

"""Callbacks that run at each training loop :class:`.Event`.

Each callback inherits from the :class:`.Callback` base class. See detailed description and
examples for writing your own callbacks at the :class:`.Callback` base class.
"""
from composer.callbacks.activation_monitor import ActivationMonitor
from composer.callbacks.checkpoint_saver import CheckpointSaver
from composer.callbacks.early_stopper import EarlyStopper
from composer.callbacks.eval_output_logging_callback import EvalOutputLogging
from composer.callbacks.export_for_inference import ExportForInferenceCallback
from composer.callbacks.free_outputs import FreeOutputs
from composer.callbacks.generate import Generate
from composer.callbacks.image_visualizer import ImageVisualizer
from composer.callbacks.lr_monitor import LRMonitor
from composer.callbacks.memory_monitor import MemoryMonitor
from composer.callbacks.memory_snapshot import MemorySnapshot
from composer.callbacks.mlperf import MLPerfCallback
from composer.callbacks.nan_monitor import NaNMonitor
from composer.callbacks.oom_observer import OOMObserver
from composer.callbacks.optimizer_monitor import OptimizerMonitor
from composer.callbacks.runtime_estimator import RuntimeEstimator
from composer.callbacks.speed_monitor import SpeedMonitor
from composer.callbacks.system_metrics_monitor import SystemMetricsMonitor
from composer.callbacks.threshold_stopper import ThresholdStopper

__all__ = [
    'ActivationMonitor',
    'OptimizerMonitor',
    'LRMonitor',
    'MemoryMonitor',
    'NaNMonitor',
    'SpeedMonitor',
    'CheckpointSaver',
    'MLPerfCallback',
    'EarlyStopper',
    'EvalOutputLogging',
    'ExportForInferenceCallback',
    'ThresholdStopper',
    'ImageVisualizer',
    'RuntimeEstimator',
    'SystemMetricsMonitor',
    'Generate',
    'FreeOutputs',
    'MemorySnapshot',
    'OOMObserver',
]



# Copyright 2022 MosaicML Composer authors
# SPDX-License-Identifier: Apache-2.0


class NoEffectWarning(Warning):
    """Warns when an algorithm did not have an effect.

    An algorithm should emit this warning when its application resulted in no changes to the trainer :class:`~.State`.
    For example, if surgery algorithms that find and replace layers find no layers to replace.
    """
    pass


class NotIntendedUseWarning(Warning):
    """Warns when an algorithm is being applied outside its intended setting.

    This is a catch-all generic warning to alert the user to potentially unintended applications. For example, when a
    data augmentation that resizes the image is being applied to a very small image, or being applied to text data.
    """
    pass



# Copyright 2024 MosaicML Composer authors
# SPDX-License-Identifier: Apache-2.0

"""Utilities for warnings."""

__all__ = ['VersionedDeprecationWarning']


class VersionedDeprecationWarning(DeprecationWarning):
    """A custom deprecation warning class that includes version information.

    Attributes:
        message (str): The deprecation message describing why the feature is deprecated.
        remove_version (str): The version in which the feature will be removed.
    """

    def __init__(self, message: str, remove_version: str) -> None:
        super().__init__(message + f' It will be removed in version {remove_version}.')



# Copyright 2022 MosaicML Composer authors
# SPDX-License-Identifier: Apache-2.0

"""Monitor learning rate during training."""
from composer.core import Callback, State
from composer.loggers import Logger

__all__ = ['LRMonitor']


class LRMonitor(Callback):
    """Logs the learning rate.

    This callback iterates over all optimizers and their parameter groups to log
    learning rate under the ``lr-{OPTIMIZER_NAME}/group{GROUP_NUMBER}`` key.

    Example:
        .. doctest::

            >>> from composer import Trainer
            >>> from composer.callbacks import LRMonitor
            >>> # constructing trainer object with this callback
            >>> trainer = Trainer(
            ...     model=model,
            ...     train_dataloader=train_dataloader,
            ...     eval_dataloader=eval_dataloader,
            ...     optimizers=optimizer,
            ...     max_duration="1ep",
            ...     callbacks=[LRMonitor()],
            ... )

    The learning rate is logged by the :class:`.Logger` to the following key as described
    below.

    +---------------------------------------------+---------------------------------------+
    | Key                                         | Logged data                           |
    +=============================================+=======================================+
    |                                             | Learning rate for each optimizer and  |
    | ``lr-{OPTIMIZER_NAME}/group{GROUP_NUMBER}`` | parameter group for that optimizer is |
    |                                             | logged to a separate key.             |
    +---------------------------------------------+---------------------------------------+
    """

    def __init__(self) -> None:
        pass

    def batch_end(self, state: State, logger: Logger):
        assert state.optimizers is not None, 'optimizers must be defined'
        for optimizer in state.optimizers:
            lrs = [group['lr'] for group in optimizer.param_groups]
            name = optimizer.__class__.__name__
            for idx, lr in enumerate(lrs):
                logger.log_metrics({f'lr-{name}/group{idx}': lr})



# Copyright 2022 MosaicML Composer authors
# SPDX-License-Identifier: Apache-2.0

"""Free train metrics."""

import torch

from composer.core import Callback, State
from composer.loggers import Logger


class FreeOutputs(Callback):
    """Free train metrics on AFTER_LOSS to reduce peak memory usage if not using train metrics."""

    def after_loss(self, state: State, logger: Logger) -> None:
        state.outputs = torch.Tensor()



# Copyright 2022 MosaicML Composer authors
# SPDX-License-Identifier: Apache-2.0

"""NoOpModel algorithm and class."""

from __future__ import annotations

import logging
from typing import TYPE_CHECKING, Any, Optional

import torch
import torch.nn.functional as F
from torchmetrics import Metric
from torchmetrics.classification import BinaryAccuracy

from composer.core import Algorithm, Event, State
from composer.loggers import Logger
from composer.models.base import ComposerModel
from composer.utils import module_surgery

if TYPE_CHECKING:
    from composer.core.types import Batch

log = logging.getLogger(__name__)

__all__ = ['NoOpModelClass', 'NoOpModel']


class NoOpModelClass(ComposerModel):
    """Dummy model used for performance measurements.

    The :class:`.NoOpModel` algorithm uses this class to replace a :class:`torch.nn.Module`.

    Args:
        original_model (torch.nn.Module): Model to replace.
    """

    def __init__(self, original_model: torch.nn.Module):
        super().__init__()
        original_device = next(original_model.parameters()).device
        self.weights = torch.nn.Parameter(torch.Tensor([1.5]).to(original_device))
        try:
            # For classification
            self.num_classes = original_model.num_classes
        except AttributeError:
            pass

    def loss(self, outputs: torch.Tensor, batch: Batch):
        x, y = batch
        assert isinstance(y, torch.Tensor)
        del x  # unused
        return F.mse_loss(outputs, y.to(torch.float32))

    def forward(self, batch: Batch):
        x, y = batch
        del x  # unused
        assert isinstance(y, torch.Tensor)
        return y * self.weights

    def get_metrics(self, is_train: bool) -> dict[str, Metric]:
        return {'BinaryAccuracy': BinaryAccuracy()}

    def eval_forward(self, batch: Batch, outputs: Optional[Any] = None):
        x, y = batch
        del x  # unused
        return y

    def update_metric(self, batch: Any, outputs: Any, metric: Metric) -> None:
        _, targets = batch
        metric.update(outputs, targets)


class NoOpModel(Algorithm):
    """Runs on :attr:`Event.INIT` and replaces the model with a dummy :class:`.NoOpModelClass` instance."""

    def __init__(self) -> None:
        # No arguments
        pass

    def __repr__(self) -> str:
        return f'{self.__class__.__name__}()'

    @staticmethod
    def required_on_load() -> bool:
        return True

    def match(self, event: Event, state: State) -> bool:
        return event == Event.INIT

    def apply(self, event: Event, state: State, logger: Logger) -> Optional[int]:
        new_model = NoOpModelClass(state.model)
        module_surgery.update_params_in_optimizer(
            old_params=state.model.parameters(),
            new_params=new_model.parameters(),
            optimizers=state.optimizers,
        )
        state.model = new_model

        log.info('Replaced model with a NoOpModel')



# Copyright 2022 MosaicML Composer authors
# SPDX-License-Identifier: Apache-2.0

"""Helper utilities for algorithms."""
from composer.algorithms.utils.augmentation_primitives import augmentation_sets

__all__ = ['augmentation_sets']
